# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import logging

from .abstract_check import ImageAbstractCheck, DockerfileAbstractCheck
from .check_utils import check_label
from ..result import CheckResult

logger = logging.getLogger(__name__)


class LabelAbstractCheck(ImageAbstractCheck, DockerfileAbstractCheck):

    def __init__(self, message, description, reference_url, tags, labels, required,
                 value_regex=None):
        """
        Abstract check for Dockerfile/Image labels.

        :param message: str
        :param description: str
        :param reference_url: str
        :param tags: [str]
        :param labels: [str]
        :param required: bool
        :param value_regex: str (using search method)
        """
        super(LabelAbstractCheck, self) \
            .__init__(message, description, reference_url, tags)
        self.labels = labels
        self.required = required
        self.value_regex = value_regex

    def check(self, target):
        passed = check_label(labels=self.labels,
                             required=self.required,
                             value_regex=self.value_regex,
                             target_labels=target.labels)

        return CheckResult(ok=passed,
                           description=self.description,
                           message=self.message,
                           reference_url=self.reference_url,
                           check_name=self.name,
                           logs=[])


class DeprecatedLabelAbstractCheck(ImageAbstractCheck, DockerfileAbstractCheck):

    def __init__(self, message, description, reference_url, tags, old_label, new_label):
        super(DeprecatedLabelAbstractCheck, self) \
            .__init__(message, description, reference_url, tags)
        self.old_label = old_label
        self.new_label = new_label

    def check(self, target):
        labels = target.labels
        old_present = labels is not None and self.old_label in labels

        passed = (not old_present) or (self.new_label in labels)

        return CheckResult(ok=passed,
                           description=self.description,
                           message=self.message,
                           reference_url=self.reference_url,
                           check_name=self.name,
                           logs=[])


class InheritedOptionalLabelAbstractCheck(ImageAbstractCheck):

    def __init__(self, message, description, reference_url, tags):
        """
        Abstract check for Dockerfile/Image labels.

        :param message: str
        :param description: str
        :param reference_url: str
        :param tags: [str]
        """
        super(InheritedOptionalLabelAbstractCheck, self) \
            .__init__(message, description, reference_url, tags)
        self.labels_list = []

    def check(self, target):
        passed = True
        logs = []

        if target.parent_target:
            labels_to_check = (set(self.labels_list) & set(target.labels)
                               & set(target.parent_target.labels))
            for label in labels_to_check:
                if target.labels[label] == target.parent_target.labels[label]:
                    passed = False
                    log = "optional label inherited: {}".format(label)
                    logs.append(log)
                    logger.debug(log)

        return CheckResult(ok=passed,
                           description=self.description,
                           message=self.message,
                           reference_url=self.reference_url,
                           check_name=self.name,
                           logs=logs)



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os

import pytest

from colin.core.exceptions import ColinRulesetException
from colin.core.ruleset.ruleset import Ruleset


def test_ruleset_yaml():
    tests_dir = os.path.dirname(os.path.dirname(__file__))
    lol_ruleset_path = os.path.join(tests_dir, "data", "lol-ruleset.yaml")
    with open(lol_ruleset_path, "r") as fd:
        r = Ruleset(ruleset_file=fd)
        checks = r.get_checks(None)
    assert len(checks) == 1


def test_ruleset_json():
    tests_dir = os.path.dirname(os.path.dirname(__file__))
    lol_ruleset_path = os.path.join(tests_dir, "data", "lol-ruleset.json")
    with open(lol_ruleset_path, "r") as fd:
        r = Ruleset(ruleset_file=fd)
        checks = r.get_checks(None)
    assert len(checks) == 1


def test_ruleset_tags():
    tags = ["a", "banana"]
    r = {
        "version": "1",
        "checks": [
            {
                "name": "name_label",
                "tags": tags[:]
            }
        ]
    }
    r = Ruleset(ruleset=r)
    checks = r.get_checks(None)
    assert len(checks) == 1
    assert checks[0].tags == tags


def test_ruleset_additional_tags():
    tags = ["a"]
    r = {
        "version": "1",
        "checks": [
            {
                "name": "name_label",
                "additional_tags": tags[:]
            }
        ]
    }
    r = Ruleset(ruleset=r)
    checks = r.get_checks(None)
    assert len(checks) == 1
    assert list(set(tags).intersection(set(checks[0].tags))) == tags


@pytest.mark.parametrize("tags,expected_check_name", [
    (["banana"], None),
    (["name"], "name_label")
])
def test_ruleset_tags_filtering(tags, expected_check_name):
    r = {
        "version": "1",
        "checks": [
            {
                "name": "name_label"
            }
        ]
    }
    r = Ruleset(ruleset=r)
    checks = r.get_checks(None, tags=tags)
    if expected_check_name:
        assert len(checks) == 1
        assert checks[0].name == expected_check_name
    else:
        assert len(checks) == 0


# version in ruleset, should this case raise an exception?
@pytest.mark.parametrize("version,should_raise", [
    (1, False),
    ("1", False),
    ("banana", True),
    (None, True),
    ("", True),
    ("<blank>", True),
])
def test_ruleset_version(version, should_raise):
    if version == "<blank>":
        r = {"banana": 123}
    else:
        r = {"version": version}
    if should_raise:
        with pytest.raises(ColinRulesetException):
            Ruleset(ruleset=r)
    else:
        assert Ruleset(ruleset=r)


def test_ruleset_override():
    m = "my-message!"
    r = {
        "version": "1",
        "checks": [
            {
                "name": "name_label",
                "tags": ["a", "b"],
                "just": "testing",
                "message": m
            }
        ]
    }
    r = Ruleset(ruleset=r)
    checks = r.get_checks(None)
    assert len(checks) == 1
    assert checks[0].message == m
    assert checks[0].just == "testing"



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import pytest
import six

import colin


@pytest.fixture()
def dockerfile_with_missing_from():
    return six.StringIO("FROMM base\nENV a=b\nLABEL c=d\n")


@pytest.fixture()
def ruleset_from_tag_not_latest():
    return {
        "version": "1",
        "name": "Laughing out loud ruleset",
        "description": "This set of checks is required to pass because we said it",
        "contact_email": "forgot-to-reply@example.nope",
        "checks": [
            {
                "name": "from_tag_not_latest"
            },
        ]
    }


def test_missing_from(dockerfile_with_missing_from,
                      ruleset_from_tag_not_latest):
    result = colin.run(target=dockerfile_with_missing_from,
                       target_type="dockerfile",
                       ruleset=ruleset_from_tag_not_latest)
    assert result
    assert not result.ok
    assert result.results_per_check['from_tag_not_latest']
    assert not result.results_per_check['from_tag_not_latest'].ok
    assert result.results_per_check['from_tag_not_latest'].status == 'ERROR'
    assert result.results_per_check['from_tag_not_latest'].logs[
               0] == 'Cannot find FROM instruction.'



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import io
import json
import logging
import os
import shutil
import subprocess
from tempfile import mkdtemp

from dockerfile_parse import DockerfileParser

from .checks.abstract_check import ImageAbstractCheck, DockerfileAbstractCheck
from ..core.exceptions import ColinException
from ..utils.cont import ImageName

logger = logging.getLogger(__name__)


def is_compatible(target_type, check_instance):
    """
    Check the target compatibility with the check instance.

    :param target_type: Target subclass, if None, returns True
    :param check_instance: instance of some Check class
    :return: True if the target is None or compatible with the check instance
    """
    if not target_type:
        return True
    return isinstance(check_instance, target_type.get_compatible_check_class())


# we've introduced an API-breaking change in conu, we need to solve this at conu level:
# https://github.com/user-cont/conu/issues/220
# in the meantime, let's workaround here
def inspect_object(obj, refresh=True):
    """
    inspect provided object (container, image) and return raw dict with the metadata

    :param obj: instance of Container or an Image
    :param refresh: bool, refresh the metadata or return cached?
    :return: dict
    """
    if hasattr(obj, "inspect"):
        return obj.inspect(refresh=refresh)
    return obj.get_metadata(refresh=refresh)


class Target(object):
    """
    Target is the thing we are going to check; it can be
    - an image (specified by name, ostree or dockertar)
    - dockerfile (specified by path or file-like object)
    """

    def __init__(self):
        self._labels = None
        self.target_name = None
        self.parent_target = None

    @property
    def labels(self):
        """
        Get list of labels from the target instance.

        :return: [str]
        """
        return None

    def clean_up(self):
        """
        Perform clean up on the low level objects: atm atomic and skopeo mountpoints
        and data are being cleaned up.
        """
        pass

    @classmethod
    def get_compatible_check_class(cls):
        """
        Get the compatible abstract check class.
        :return: cls
        """
        return None

    @staticmethod
    def get_instance(target_type, **kwargs):
        """
        :param target_type: string, either image, dockertar, ostree or dockerfile
        """
        if target_type in TARGET_TYPES:
            cls = TARGET_TYPES[target_type]
            try:
                return cls(**kwargs)
            except Exception:
                logger.error("Please make sure that you picked the correct target type: "
                             "--target-type CLI option.")
                raise

        raise ColinException(
            "Unknown target type '{}'. Please make sure that you picked the correct target type: "
            "--target-type CLI option.".format(target_type))


class DockerfileTarget(Target):
    target_type = "dockerfile"

    def __init__(self, target, **_):
        super().__init__()
        self.target_name = target
        logger.debug("Target is a dockerfile.")
        if isinstance(target, io.IOBase):
            logger.debug("Target is a dockerfile loaded from the file-like object.")
            self.instance = DockerfileParser(fileobj=target)
        else:
            self.instance = DockerfileParser(fileobj=open(target))

    @property
    def labels(self):
        """
        Get list of labels from the target instance.

        :return: [str]
        """
        if self._labels is None:
            self._labels = self.instance.labels
        return self._labels

    @classmethod
    def get_compatible_check_class(cls):
        return DockerfileAbstractCheck


class AbstractImageTarget(Target):
    """
    Abstract predecessor for the image-target classes. (e.g. ostree and podman image)
    """

    @property
    def config_metadata(self):
        """ metadata from "Config" key """
        raise NotImplementedError("Unsupported right now.")

    @property
    def mount_point(self):
        """ real filesystem """
        raise NotImplementedError("Unsupported right now.")

    def get_output(self, cmd):
        """
        Get output of the command from the container based on this image.
        :param cmd: [str]
        :return: str
        """
        raise NotImplementedError("Unsupported right now.")

    def read_file(self, file_path):
        """
        read file specified via 'file_path' and return its content - raises an ConuException if
        there is an issue accessing the file
        :param file_path: str, path to the file to read
        :return: str (not bytes), content of the file
        """
        try:
            with open(self.cont_path(file_path)) as fd:
                return fd.read()
        except IOError as ex:
            logger.error("error while accessing file %s: %r", file_path, ex)
            raise ColinException(
                "There was an error while accessing file %s: %r" % (file_path, ex))

    def get_file(self, file_path, mode="r"):
        """
        provide File object specified via 'file_path'
        :param file_path: str, path to the file
        :param mode: str, mode used when opening the file
        :return: File instance
        """
        return open(self.cont_path(file_path), mode=mode)

    def file_is_present(self, file_path):
        """
        check if file 'file_path' is present, raises IOError if file_path
        is not a file
        :param file_path: str, path to the file
        :return: True if file exists, False if file does not exist
        """
        real_path = self.cont_path(file_path)
        if not os.path.exists(real_path):
            return False
        if not os.path.isfile(real_path):
            raise IOError("%s is not a file" % file_path)
        return True

    def cont_path(self, path):
        """
        provide absolute path within the container

        :param path: path with container
        :return: str
        """
        if path.startswith("/"):
            path = path[1:]
        real_path = os.path.join(self.mount_point, path)
        logger.debug("path = %s", real_path)
        return real_path

    @classmethod
    def get_compatible_check_class(cls):
        return ImageAbstractCheck


class ImageTarget(AbstractImageTarget):
    """
    Represents the podman image as a target.
    """
    target_type = "image"

    def __init__(self, target, pull, parent_target=None, insecure=False, **_):
        super().__init__()
        logger.debug("Target is an image.")
        self.pull = pull
        self.insecure = insecure
        self.image_name_obj = ImageName.parse(target)
        self.target_name = self.image_name_obj.name
        self.parent_target = parent_target

        self._config_metadata = None
        self._mount_point = None
        self._mounted_container_id = None
        self.image_id = None

        self._try_image()

    @property
    def config_metadata(self):
        if not self._config_metadata:
            cmd = ["podman", "inspect", self.target_name]
            loaded_config = json.loads(subprocess.check_output(cmd))
            if loaded_config and isinstance(loaded_config, list):
                self._config_metadata = loaded_config[0]
                # FIXME: Better validation.
            else:
                raise ColinException("Cannot load config for the image.")

        return self._config_metadata

    @property
    def labels(self):
        return self.config_metadata["Labels"] or {}

    @property
    def mount_point(self):
        """ podman mount -- real filesystem """
        if self._mount_point is None:
            cmd_create = ["podman", "create", self.target_name, "some-cmd"]
            self._mounted_container_id = subprocess.check_output(cmd_create).decode().rstrip()
            cmd_mount = ["podman", "mount", self._mounted_container_id]
            self._mount_point = subprocess.check_output(cmd_mount).decode().rstrip()
        return self._mount_point

    def _try_image(self):
        logger.debug("Trying to find an image.")
        cmd = ["podman", "images", "--quiet", self.target_name]
        result = subprocess.run(cmd,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        if result.returncode == 0:
            self.image_id = result.stdout.decode().rstrip()
            logger.debug("Image found with id: '{}'.".format(self.image_id))
        else:
            if "unable to find" in result.stderr.decode():
                if self.pull:
                    logger.debug("Pulling an image.")
                    cmd_pull = ["podman", "pull", "--quiet", self.target_name]
                    result_pull = subprocess.run(cmd_pull,
                                                 stdout=subprocess.PIPE,
                                                 stderr=subprocess.PIPE)
                    if result_pull.returncode == 0:
                        self.image_id = result_pull.stdout.decode().rstrip()
                        logger.debug("Image pulled with id: '{}'.".format(self.image_id))
                    else:
                        raise ColinException(
                            "Cannot pull an image: '{}'.".format(self.target_name))

                else:
                    raise ColinException("Image '{}' not found.".format(self.target_name))
            else:
                raise ColinException("Podman error: {}".format(result.stderr))

    def clean_up(self):
        if self._mount_point:
            cmd = ["podman", "umount", self._mounted_container_id]
            subprocess.check_call(cmd, stdout=subprocess.DEVNULL)
            self._mount_point = None
        if self._mounted_container_id:
            cmd = ["podman", "rm", self._mounted_container_id]
            subprocess.check_call(cmd, stdout=subprocess.DEVNULL)
            self._mounted_container_id = None

    def get_output(self, cmd):
        raise NotImplementedError("Unsupported right now.")


class OstreeTarget(AbstractImageTarget):
    """
    Represents the ostree repository as an image target.
    """
    target_type = "ostree"

    def __init__(self, target, parent_target=None, **_):
        super().__init__()
        logger.debug("Target is an ostree repository.")

        self.target_name = target
        if self.target_name.startswith("ostree:"):
            self.target_name = self.target_name[7:]

        try:
            self.ref_image_name, self._ostree_path = self.target_name.split("@", 1)
        except ValueError:
            raise RuntimeError("Invalid ostree target: should be 'image@path'.")

        self.parent_target = parent_target
        self._tmpdir = None
        self._mount_point = None
        self._layers_path = None
        self._labels = None

    @property
    def labels(self):
        """
        Provide labels without the need of dockerd. Instead skopeo is being used.

        :return: dict
        """
        if self._labels is None:
            cmd = ["skopeo", "inspect", self.skopeo_target]
            self._labels = json.loads(subprocess.check_output(cmd))["Labels"]
        return self._labels

    @property
    def layers_path(self):
        """ Directory with all the layers (docker save). """
        if self._layers_path is None:
            self._layers_path = os.path.join(self.tmpdir, "layers")
        return self._layers_path

    @property
    def mount_point(self):
        """ ostree checkout -- real filesystem """
        if self._mount_point is None:
            self._mount_point = os.path.join(self.tmpdir, "checkout")
            os.makedirs(self._mount_point)
            self._checkout()
        return self._mount_point

    @property
    def ostree_path(self):
        """ ostree repository -- content """
        if self._ostree_path is None:
            self._ostree_path = os.path.join(self.tmpdir, "ostree-repo")
            subprocess.check_call(["ostree", "init", "--mode", "bare-user-only",
                                   "--repo", self._ostree_path])
        return self._ostree_path

    @property
    def skopeo_target(self):
        """ Skopeo format for the ostree repository. """
        return "ostree:{}@{}".format(self.ref_image_name, self.ostree_path)

    @property
    def tmpdir(self):
        """ Temporary directory holding all the runtime data. """
        if self._tmpdir is None:
            self._tmpdir = mkdtemp(prefix="colin-", dir="/var/tmp")
        return self._tmpdir

    def clean_up(self):
        cmd = ["atomic", "unmount", self.mount_point]
        self._run_and_log(cmd, self.ostree_path, "Failed to unmount ostree checkout.")
        shutil.rmtree(self.tmpdir)

    def _checkout(self):
        """ check out the image filesystem on self.mount_point """
        cmd = ["atomic", "mount", "--storage", "ostree", self.ref_image_name, self.mount_point]
        # self.mount_point has to be created by us
        self._run_and_log(cmd, self.ostree_path,
                          "Failed to mount selected image as an ostree repo.")

    @staticmethod
    def _run_and_log(cmd, ostree_repo_path, error_msg, wd=None):
        """ run provided command and log all of its output; set path to ostree repo """
        logger.debug("running command %s", cmd)
        kwargs = {
            "stderr": subprocess.STDOUT,
            "env": os.environ.copy(),
        }
        if ostree_repo_path:
            # must not exist, ostree will create it
            kwargs["env"]["ATOMIC_OSTREE_REPO"] = ostree_repo_path
        if wd:
            kwargs["cwd"] = wd
        try:
            out = subprocess.check_output(cmd, **kwargs)
        except subprocess.CalledProcessError as ex:
            logger.error(ex.output)
            logger.error(error_msg)
            raise
        logger.debug("%s", out)

    @property
    def config_metadata(self):
        """ metadata from "Config" key """
        raise NotImplementedError("Skopeo does not provide metadata yet.")

    def get_output(self, cmd):
        raise NotImplementedError("Unsupported right now.")


TARGET_TYPES = {
    "image": ImageTarget,
    "dockerfile": DockerfileTarget,
    "ostree": OstreeTarget
}



import logging
import os
import shutil
import subprocess
import tempfile

import pytest

from colin.core.colin import _set_logging
from colin.core.target import ImageTarget, OstreeTarget, DockerfileTarget

_set_logging(level=logging.DEBUG)

BASH_IMAGE = "colin-test-bash"
LS_IMAGE = "colin-test-ls"
BUSYBOX_IMAGE = "busybox:latest"
LABELS_IMAGE = "colin-labels"
LABELS_IMAGE_PARENT = "colin-labels-parent"
IMAGES = {
    BASH_IMAGE: {
        "dockerfile_path": "Dockerfile-bash"
    },
    LS_IMAGE: {
        "dockerfile_path": "Dockerfile-ls"
    },
    LABELS_IMAGE: {
        "dockerfile_path": "Dockerfile"
    },
    LABELS_IMAGE_PARENT: {
        "dockerfile_path": "Dockerfile-parent"
    }
}


def build_image_if_not_exists(image_name):
    try:
        subprocess.check_call(["podman", "image", "exists", image_name])
    except subprocess.CalledProcessError:
        this_dir = os.path.abspath(os.path.dirname(__file__))
        data_dir = os.path.join(this_dir, "data")

        dockerfile_path = IMAGES[image_name]["dockerfile_path"]
        cmd_create = ["podman", "build", "-t", image_name, "-f", dockerfile_path, data_dir]
        subprocess.check_call(cmd_create)


def pull_image_if_not_exists(image_name):
    try:
        subprocess.check_call(["podman", "image", "exists", image_name])
    except subprocess.CalledProcessError:
        subprocess.check_call(["podman", "pull", image_name])


def convert_image_to_ostree(image_name):
    # /tmp is tmpfs and ostree can't do its magic there
    tmpdir_path = tempfile.mkdtemp(prefix="pytest-", dir="/var/tmp")
    ostree_path = os.path.join(tmpdir_path, "os3")
    os.makedirs(ostree_path)
    skopeo_target = get_skopeo_path(image_name=image_name, ostree_path=ostree_path)

    subprocess.check_call(["ostree", "init", "--mode", "bare-user-only",
                           "--repo", ostree_path])

    cmd = ["podman", "push", image_name, skopeo_target]
    subprocess.check_call(cmd)
    return ostree_path


def get_target(name, type):
    if type == "image":

        target = ImageTarget(target=name,
                             pull=False)
        yield target
        target.clean_up()

    elif type == "ostree":

        ostree_path = convert_image_to_ostree(name)
        skopeo_target = get_skopeo_path(image_name=name,
                                        ostree_path=ostree_path)

        ostree_target = OstreeTarget(target=skopeo_target)
        yield ostree_target
        ostree_target.clean_up()
        shutil.rmtree(ostree_path)

    elif type == "dockerfile":

        this_dir = os.path.abspath(os.path.dirname(__file__))
        data_dir = os.path.join(this_dir, "data")
        dockerfile_path = os.path.join(data_dir, IMAGES[name]["dockerfile_path"])

        yield DockerfileTarget(target=dockerfile_path)


def get_skopeo_path(image_name, ostree_path):
    return "ostree:%s@%s" % (image_name, ostree_path)


@pytest.fixture(scope="session", autouse=True)
def label_image():
    build_image_if_not_exists(LABELS_IMAGE)


@pytest.fixture(scope="session",
                params=["image", "ostree", "dockerfile"])
def target_label(request, label_image):
    for t in get_target(name=LABELS_IMAGE,
                        type=request.param):
        yield t


@pytest.fixture(scope="session",
                params=["image", "ostree", "dockerfile"])
def target_label_image_and_dockerfile(request, label_image):
    for t in get_target(name=LABELS_IMAGE,
                        type=request.param):
        yield t


@pytest.fixture(scope="session", autouse=True)
def target_bash_image():
    build_image_if_not_exists(BASH_IMAGE)


@pytest.fixture(scope="session",
                params=["image", "ostree"])
def target_bash(request, target_bash_image):
    for t in get_target(name=BASH_IMAGE,
                        type=request.param):
        yield t


@pytest.fixture(scope="session", autouse=True)
def target_ls_image():
    build_image_if_not_exists(LS_IMAGE)


@pytest.fixture(scope="session",
                params=["image", "ostree"])
def target_ls(request, target_ls_image):
    for t in get_target(name=LS_IMAGE,
                        type=request.param):
        yield t


@pytest.fixture(scope="session",
                params=[LS_IMAGE, BASH_IMAGE])
def target_help_file(request, target_ls, target_bash):
    if request.param == LS_IMAGE:
        return target_ls, False
    if request.param == BASH_IMAGE:
        return target_bash, True


@pytest.fixture(scope="session", autouse=True)
def target_busybox_image():
    pull_image_if_not_exists(image_name=BUSYBOX_IMAGE)


@pytest.fixture(scope="session",
                params=["image", "ostree"])
def target_busybox(request, target_busybox_image):
    for t in get_target(name=BUSYBOX_IMAGE,
                        type=request.param):
        yield t



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import json

from six import iteritems


class AbstractCheck(object):
    name = None
    check_type = None

    def __init__(self, message, description, reference_url, tags):
        self.message = message
        self.description = description
        self.reference_url = reference_url
        self.tags = tags
        self.timeout = None

    def check(self, target):
        pass

    def __str__(self):
        return "{}\n" \
               "   -> {}\n" \
               "   -> {}\n" \
               "   -> {}\n" \
               "   -> {}\n".format(self.name,
                                   self.message,
                                   self.description,
                                   self.reference_url,
                                   ", ".join(self.tags))

    @property
    def json(self):
        """
        Get json representation of the check

        :return: dict (str -> obj)
        """
        return {
            'name': self.name,
            'message': self.message,
            'description': self.description,
            'reference_url': self.reference_url,
            'tags': self.tags,
        }

    @staticmethod
    def json_from_all_checks(checks):
        result_json = {}
        for (group, group_checks) in iteritems(checks):

            result_list = []
            for check in group_checks:
                result_list.append(check.json)

            result_json[group] = result_list
        return result_json

    @staticmethod
    def save_checks_to_json(file, checks):
        json.dump(obj=AbstractCheck.json_from_all_checks(checks=checks),
                  fp=file,
                  indent=4)


class DockerfileAbstractCheck(AbstractCheck):
    check_type = "dockerfile"


class ImageAbstractCheck(AbstractCheck):
    check_type = "image"


class FilesystemAbstractCheck(AbstractCheck):
    pass



"""
Test different target types.
"""

import pytest

import colin
from tests.conftest import LABELS_IMAGE, convert_image_to_ostree, get_skopeo_path


@pytest.fixture()
def ruleset():
    """ simple ruleset as a pytest fixture """
    return {
        "version": "1",
        "name": "Laughing out loud ruleset",
        "description": "This set of checks is required to pass because we said it",
        "contact_email": "forgot-to-reply@example.nope",
        "checks": [
            {
                "name": "com.redhat.component_label",
            },
            {
                "name": "name_label",
            },
            {
                "name": "version_label",
            },
            {
                "name": "description_label",
            },
            {
                "name": "io.k8s.description_label",
            },
            {
                "name": "vcs-ref_label",
            },
            {
                "name": "vcs-type_label",
            },
            {
                "name": "architecture_label",
            },
            {
                "name": "com.redhat.build-host_label",
            },
            {
                "name": "authoritative_source-url_label",
            },
            {
                "name": "vendor_label",
            },
            {
                "name": "release_label",
            },
            {
                "name": "url_label",
                "usable_targets": ["image"]
            },
            {
                "name": "build-date_label",
            },
            {
                "name": "distribution-scope_label",
            },
            {
                "name": "run_or_usage_label",
            },
            {
                "name": "help_file_or_readme",
            },
            {
                "name": "maintainer_label",
            },
            {
                "name": "summary_label",
            },
            {
                "name": "install_label_capital_deprecated",
            },
            {
                "name": "architecture_label_capital_deprecated",
            },
            {
                "name": "bzcomponent_deprecated",
            },
            {
                "name": "name_label_capital_deprecated",
            },
            {
                "name": "release_label_capital_deprecated",
            },
            {
                "name": "uninstall_label_capital_deprecated",
            },
            {
                "name": "version_label_capital_deprecated",
            },

        ]
    }


def test_podman_image_target(ruleset):
    results = colin.run(LABELS_IMAGE, "image", ruleset=ruleset, logging_level=10, pull=False)
    assert results.ok
    assert results.results_per_check["url_label"].ok


def test_ostree_target(ruleset):
    image_name = "colin-labels"
    ostree_path = convert_image_to_ostree(image_name=image_name)
    skopeo_target = get_skopeo_path(image_name=image_name, ostree_path=ostree_path)
    results = colin.run(skopeo_target, "ostree", ruleset=ruleset, logging_level=10, pull=False)
    assert results.ok
    assert results.results_per_check["url_label"].ok



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from colin.core.checks.abstract_check import AbstractCheck


class FunkyCheck(AbstractCheck):
    name = "this-is-a-funky-check"

    def __init__(self):
        super(FunkyCheck, self).__init__(
            message="yes!",
            description="no",
            reference_url="https://nope.example.com/",
            tags=["yes", "and", "no"]
        )


class ThisIsNotAChekk(AbstractCheck):
    name = "this-is-not-a-check"

    def __init__(self):
        super(ThisIsNotAChekk, self).__init__(
            message="yes!",
            description="no",
            reference_url="https://nope.example.com/",
            tags=["yes", "and", "no"]
        )


class ThisIsAlsoNotAChekk(object):
    name = "this-is-also-not-a-check"

    def __init__(self):
        super(ThisIsAlsoNotAChekk, self).__init__(
            message="yes!",
            description="no",
            reference_url="https://nope.example.com/",
            tags=["yes", "and", "no"]
        )



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from click.testing import CliRunner

from colin.cli.colin import check, list_checks, list_rulesets, info
from colin.version import __version__


def _call_colin(fnc, parameters=None, envs=None):
    runner = CliRunner()
    envs = envs or {}
    if not parameters:
        return runner.invoke(fnc, env=envs)
    else:
        return runner.invoke(fnc, parameters, env=envs)


def _common_help_options(result):
    assert "-r, --ruleset TEXT" in result.output
    assert "-f, --ruleset-file FILENAME" in result.output
    assert "--debug" in result.output
    assert "--json FILENAME" in result.output
    assert "-t, --tag TEXT" in result.output
    assert "-v, --verbose" in result.output
    assert "-h, --help" in result.output


def test_check_command():
    result = _call_colin(check)
    expected_output1 = "Usage: check [OPTIONS] TARGET"
    expected_output2 = 'Error: Missing argument "TARGET"'
    assert result.exit_code == 2
    assert expected_output1 in result.output
    assert expected_output2 in result.output


def test_check_help_command():
    result = _call_colin(check, parameters=["-h"])
    assert result.exit_code == 0
    _common_help_options(result)
    assert "--stat" in result.output


def test_list_checks():
    result = _call_colin(list_checks)
    expected_output = """maintainer_label
   -> Label 'maintainer' has to be specified.
   -> The name and email of the maintainer (usually the submitter).
   -> https://fedoraproject.org/wiki/Container:Guidelines#LABELS
   -> label, maintainer, required

from_tag_not_latest
   -> In FROM, tag has to be specified and not 'latest'.
   -> Using the 'latest' tag may cause unpredictable builds.It is recommended that a specific tag is used in the FROM.
   -> https://fedoraproject.org/wiki/Container:Guidelines#FROM
   -> dockerfile, from, baseimage, latest, required

maintainer_deprecated
   -> Dockerfile instruction `MAINTAINER` is deprecated.
   -> Replace with label 'maintainer'.
   -> https://docs.docker.com/engine/reference/builder/#maintainer-deprecated
   -> dockerfile, maintainer, deprecated, required

"""
    assert result.exit_code == 0
    assert result.output == expected_output
    assert _call_colin(list_checks, parameters=["-r", "default"]).output == \
           _call_colin(list_checks).output


def test_list_checks_help_command():
    result = _call_colin(list_checks, parameters=["-h"])
    _common_help_options(result)
    assert result.exit_code == 0


def test_list_checks_fedora():
    result = _call_colin(list_checks, parameters=["-r", "fedora"])
    assert result.exit_code == 0
    assert "maintainer_label" in result.output
    assert "maintainer_deprecated" in result.output
    assert "from_tag_not_latest" in result.output


def test_list_rulesets():
    result = _call_colin(list_rulesets)
    assert "fedora" in result.output
    assert "default" in result.output
    assert result.exit_code == 0


def test_list_rulesets_help_command():
    result = _call_colin(list_rulesets, parameters=["-h"])
    expected_result = """Usage: list-rulesets [OPTIONS]

  List available rulesets.

Options:
  --debug     Enable debugging mode (debugging logs, full tracebacks).
  -h, --help  Show this message and exit.
"""
    assert result.exit_code == 0
    assert result.output == expected_result


def test_info():
    result = _call_colin(info)
    assert result.exit_code == 0

    output = result.output.split("\n")
    assert output[0].startswith("colin")
    assert output[0].endswith("/colin")
    assert __version__ in output[0]
    assert "cli/colin.py" in output[1]

    assert output[3].startswith("podman")
    assert output[4].startswith("skopeo")
    assert output[5].startswith("ostree")


def test_env():
    result = _call_colin(info, envs={"COLIN_HELP": "1"})
    assert "Usage" in result.output



import itertools


class CachingIterable(object):
    def __init__(self, iterable):
        self.iterable = iterable
        self.iter = iter(iterable)
        self.done = False
        self.vals = []

    def __iter__(self):
        if self.done:
            return iter(self.vals)
        return itertools.chain(self.vals, self._gen_iter())

    def _gen_iter(self):
        for new_val in self.iter:
            self.vals.append(new_val)
            yield new_val
        self.done = True



"""
Module handling FMF stored metadata for classes
"""

import logging
import re

from fmf import Tree

logger = logging.getLogger(__name__)


class ExtendedTree(Tree):
    """
    FMF Extension. Allows to use references via @ to another items -> usefull for rulesets
    """

    def __remove_append_items(self, whole=False):
        """
        internal method, delete all append items (ends with +)
        :param whole: pass thru 'whole' param to climb
        :return: None
        """
        for node in self.climb(whole=whole):
            for key in sorted(node.data.keys()):
                if key.endswith('+'):
                    del node.data[key]

    def references(self, datatrees, whole=False):
        """
        Reference name resolver (eg. /a/b/c/d@.x.y or /a/b/c/@y will search data in .x.y or y nodes)
        there are used regular expressions (re.search) to match names
        it uses simple references schema, do not use references to another references,
        avoid usind / in reference because actual solution creates also these tree items.

        datatree contains for example data like (original check data)
        /dockerfile/maintainer_check:
          class: SomeClass
          tags: [dockerfile]

        and reference could be like (ruleset)
        /default/check1@maintainer_check:
           tags+: [required]

        will produce output (output ruleset tree):
        /default/check1@maintainer_check:
          class: SomeClass
          tags: [dockerfile, required]


        :param whole: 'whole' param of original climb method, in colin this is not used anyhow now
                      iterate over all items not only leaves if True
        :param datatrees: list of original trees with testcases to contain parent nodes
        :return: None
        """
        if not isinstance(datatrees, list):
            raise ValueError("datatrees argument has to be list of fmf trees")
        reference_nodes = self.prune(whole=whole, names=["@"])
        for node in reference_nodes:
            node.data = node.original_data
            ref_item_name = node.name.rsplit("@", 1)[1]
            # match item what does not contain @ before name, otherwise it
            # match same item
            reference_node = None
            for datatree in datatrees:
                reference_node = datatree.search("[^@]%s" % ref_item_name)
                if reference_node is not None:
                    break
            if not reference_node:
                raise ValueError("Unable to find reference for node: %s via name search: %s" %
                                 (node.name, ref_item_name))
            logger.debug("MERGING: %s @ %s from %s",
                         node.name,
                         reference_node.name,
                         reference_node.root)
            node.merge(parent=reference_node)

        self.__remove_append_items(whole=whole)

    def search(self, name):
        """ Search node with given name based on regexp, basic method (find) uses equality"""
        for node in self.climb():
            if re.search(name, node.name):
                return node
        return None



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import tempfile

import pytest
import yaml
from colin.core.exceptions import ColinRulesetException

import colin
from colin.core.checks.check_utils import NotLoadedCheck
from colin.core.checks.labels import LabelAbstractCheck


@pytest.fixture()
def ruleset():
    return {
        "version": "1",
        "name": "Laughing out loud ruleset",
        "description": "This set of checks is required to pass because we said it",
        "contact_email": "forgot-to-reply@example.nope",
        "checks": [
            {
                "name": "maintainer_label"
            },
            {
                "name": "name_label"
            },
            {
                "name": "com.redhat.component_label"
            },
            {
                "name": "help_label"
            }
        ]
    }


@pytest.fixture()
def ruleset_unknown_check():
    return {
        "version": "1",
        "name": "Laughing out loud ruleset",
        "description": "This set of checks is required to pass because we said it",
        "contact_email": "forgot-to-reply@example.nope",
        "checks": [
            {
                "name": "maintainer_label"
            },
            {
                "name": "i_forgot_the_name"
            }
        ]
    }


@pytest.fixture()
def ruleset_coupled():
    return {
        "version": "1",
        "name": "Laughing out loud coublet ruleset",
        "description": "This set of checks is required to pass because we said it",
        "contact_email": "forgot-to-reply@example.nope",
        "checks": [
            {
                "names": ["maintainer_label", "name_label", "com.redhat.component_label"],
                "additional_tags": [
                    "required"
                ]
            }
        ]
    }


@pytest.fixture()
def expected_dict():
    return {"maintainer_label": "PASS",
            "name_label": "PASS",
            "com.redhat.component_label": "PASS",
            "help_label": "FAIL",
            }


def get_results_from_colin_labels_image(image, ruleset_name=None, ruleset_file=None, ruleset=None):
    return colin.run(image.target_name, image.target_type, ruleset_name=ruleset_name,
                     ruleset_file=ruleset_file, ruleset=ruleset)


def test_specific_ruleset_as_fileobj(tmpdir, ruleset, expected_dict,
                                     target_label):
    (_, t) = tempfile.mkstemp(dir=str(tmpdir))

    with open(t, "w") as f:
        yaml.dump(ruleset, f)
    with open(t, "r") as f:
        result = get_results_from_colin_labels_image(image=target_label,
                                                     ruleset_file=f)
    assert result
    labels_dict = {}
    for res in result.results:
        labels_dict[res.check_name] = res.status
    for key in expected_dict.keys():
        assert labels_dict[key] == expected_dict[key]


def test_specific_ruleset_directly(ruleset, expected_dict, target_label):
    result = get_results_from_colin_labels_image(image=target_label,
                                                 ruleset=ruleset)
    assert result
    labels_dict = {}
    for res in result.results:
        labels_dict[res.check_name] = res.status
    for key in expected_dict.keys():
        assert labels_dict[key] == expected_dict[key]


def test_get_checks_directly(ruleset):
    checks = colin.get_checks(ruleset=ruleset)
    assert checks


def test_coupled_ruleset(ruleset_coupled):
    checks = colin.get_checks(ruleset=ruleset_coupled)
    assert checks
    assert len(checks) == 3
    for c in checks:
        assert "required" in c.tags


def test_unknown_check(ruleset_unknown_check):
    with pytest.raises(ColinRulesetException) as ex:
        colin.get_checks(ruleset=ruleset_unknown_check)
    assert str(ex.value) == "Check i_forgot_the_name can't be loaded, we couldn't find it."


def test_skip(ruleset):
    checks = colin.get_checks(ruleset=ruleset, skips=["name_label", "help_label"])
    assert len(checks) == 2
    for check in checks:
        assert check.name in ["com.redhat.component_label", "maintainer_label"]



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import pytest


@pytest.fixture()
def ruleset():
    return {
        "version": "1",
        "name": "Laughing out loud ruleset",
        "description": "This set of checks is required to pass because we said it",
        "contact_email": "forgot-to-reply@example.nope",
        "checks": [
            {
                "name": "shell_runnable"
            }
        ]
    }

# def test_dynamic_check_ls(ruleset):
#     results = colin.run(target=LS_IMAGE, ruleset=ruleset, logging_level=10)
#     assert not results.ok
#
#
# def test_dynamic_check_bash(ruleset):
#     results = colin.run(target=BASH_IMAGE, ruleset=ruleset, logging_level=10)
#     assert results.ok



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import pytest

import colin


@pytest.fixture()
def ruleset():
    """ simple ruleset as a pytest fixture """
    return {
        "version": "1",
        "name": "Laughing out loud ruleset",
        "description": "This set of checks is required to pass because we said it",
        "contact_email": "forgot-to-reply@example.nope",
        "checks": [
            {
                "name": "help_file_or_readme"
            }
        ]
    }


def test_help_file_or_readme_bash(ruleset, target_bash):
    help_file_or_readme_test(ruleset=ruleset,
                             image=target_bash,
                             should_pass=True)


def test_help_file_or_readme_ls(ruleset, target_ls):
    help_file_or_readme_test(ruleset=ruleset,
                             image=target_ls,
                             should_pass=False)


def help_file_or_readme_test(ruleset, image, should_pass):
    """ verify that help_file_or_readme check works well """
    results = colin.run(target=image.target_name,
                        target_type=image.target_type,
                        ruleset=ruleset, logging_level=10, pull=False)
    assert results.ok
    assert results.fail is not should_pass



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import json

import six

from .constant import (COLOURS, ERROR, FAILED, OUTPUT_CHARS, PASSED)
from ..utils.caching_iterable import CachingIterable


class CheckResult(object):

    def __init__(self, ok, description, message, reference_url, check_name, logs):
        self.ok = ok
        self.description = description
        self.message = message
        self.reference_url = reference_url
        self.check_name = check_name
        self.logs = logs

    @property
    def status(self):
        return PASSED if self.ok else FAILED

    def __str__(self):
        return "{}:{}".format(self.status,
                              self.message)


class DockerfileCheckResult(CheckResult):

    def __init__(self, ok, description, message, reference_url, check_name, lines=None,
                 correction_diff=None):
        super(DockerfileCheckResult, self) \
            .__init__(ok, description, message, reference_url, check_name)
        self.lines = lines
        self.correction_diff = correction_diff


class CheckResults(object):

    def __init__(self, results):
        self.results = CachingIterable(results)

    @property
    def results_per_check(self):
        return {r.check_name: r for r in self.results}

    @property
    def _dict_of_results(self):
        """
        Get the dictionary representation of results

        :return: dict (str -> dict (str -> str))
        """
        result_json = {}

        result_list = []
        for r in self.results:
            result_list.append({
                'name': r.check_name,
                'ok': r.ok,
                'status': r.status,
                'description': r.description,
                'message': r.message,
                'reference_url': r.reference_url,
                'logs': r.logs,
            })
        result_json["checks"] = result_list
        return result_json

    @property
    def json(self):
        """
        Get the json representation of results

        :return: str
        """
        return json.dumps(self._dict_of_results, indent=4)

    def save_json_to_file(self, file):
        json.dump(obj=self._dict_of_results,
                  fp=file,
                  indent=4)

    @property
    def statistics(self):
        """
        Get the dictionary with the count of the check-statuses

        :return: dict(str -> int)
        """
        result = {}
        for r in self.results:
            result.setdefault(r.status, 0)
            result[r.status] += 1
        return result

    @property
    def ok(self):
        """
        If the results ended without any error


        :return: True, if there is no check which ends with error status
        """
        return ERROR not in self.statistics

    @property
    def fail(self):
        """
        If the results ended without any fail


        :return: True, if there is no check which ends with fail status
        """
        return FAILED in self.statistics

    def generate_pretty_output(self, stat, verbose, output_function, logs=True):
        """
        Send the formated to the provided function

        :param stat: if True print stat instead of full output
        :param verbose: bool
        :param output_function: function to send output to
        """

        has_check = False
        for r in self.results:
            has_check = True
            if stat:
                output_function(OUTPUT_CHARS[r.status],
                                fg=COLOURS[r.status],
                                nl=False)
            else:
                output_function(str(r), fg=COLOURS[r.status])
                if verbose:
                    output_function("  -> {}\n"
                                    "  -> {}".format(r.description,
                                                     r.reference_url),
                                    fg=COLOURS[r.status])
                    if logs and r.logs:
                        output_function("  -> logs:",
                                        fg=COLOURS[r.status])
                        for l in r.logs:
                            output_function("    -> {}".format(l),
                                            fg=COLOURS[r.status])

        if not has_check:
            output_function("No check found.")
        elif stat and not verbose:
            output_function("")
        else:
            output_function("")
            for status, count in six.iteritems(self.statistics):
                output_function("{}:{} ".format(status, count), nl=False)
            output_function("")

    def get_pretty_string(self, stat, verbose):
        """
        Pretty string representation of the results

        :param stat: bool
        :param verbose: bool
        :return: str
        """
        pretty_output = _PrettyOutputToStr()
        self.generate_pretty_output(stat=stat,
                                    verbose=verbose,
                                    output_function=pretty_output.save_output)
        return pretty_output.result


class FailedCheckResult(CheckResult):

    def __init__(self, check, logs=None):
        super(FailedCheckResult, self) \
            .__init__(ok=False,
                      message=check.message,
                      description=check.description,
                      reference_url=check.reference_url,
                      check_name=check.name,
                      logs=logs or []
                      )

    @property
    def status(self):
        return ERROR


class _PrettyOutputToStr(object):

    def __init__(self):
        self.result = ""

    def save_output(self, text=None, fg=None, nl=True):
        text = text or ""
        self.result += text
        if nl:
            self.result += "\n"



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from colin.core.checks.labels import LabelAbstractCheck, InheritedOptionalLabelAbstractCheck
from colin.core.checks.fmf_check import FMFAbstractCheck


class ArchitectureLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "architecture_label"


class AuthoritativeSourceUrlLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "authoritative_source-url_label"


class BuildDateLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "build-date_label"
    # TODO: Check the RFC 3339 date-time format


class BuildHostLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "com.redhat.build-host_label"


class ComRedhatComponentLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "com.redhat.component_label"
    # TODO: Check the format


class DescriptionLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "description_label"


class DescriptionOrIoK8sDescriptionLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "description_or_io.k8s.description_label"


class DistributionScopeLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "distribution-scope_label"


class HelpLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "help_label"


class IoK8sDescriptionLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "io.k8s.description_label"


class IoK8sDisplayNameLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "io.k8s.display-name_label"


class MaintainerLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "maintainer_label"


class NameLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "name_label"


class ReleaseLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "release_label"


class SummaryLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "summary_label"


class UrlLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "url_label"


class RunOrUsageLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "run_or_usage_label"


class VcsRefLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "vcs-ref_label"


class VcsTypeLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "vcs-type_label"


class VcsUrlLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "vcs-url_label"


class VendorLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "vendor_label"


class VersionLabelCheck(FMFAbstractCheck, LabelAbstractCheck):
    name = "version_label"


class InheritedOptionalLabelCheck(FMFAbstractCheck, InheritedOptionalLabelAbstractCheck):
    name = "inherited_labels"



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os
import sys

import click
import six

from colin.utils.cmd_tools import get_version_msg_from_the_cmd, is_rpm_installed
from .default_group import DefaultGroup
from ..core.checks.abstract_check import AbstractCheck
from ..core.colin import get_checks, run
from ..core.constant import COLIN_CHECKS_PATH
from ..core.exceptions import ColinException
from ..core.ruleset.ruleset import get_rulesets, get_checks_paths
from ..version import __version__

logger = logging.getLogger("colin.cli")

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'],
                        auto_envvar_prefix='COLIN')


@click.group(cls=DefaultGroup, context_settings=CONTEXT_SETTINGS)
@click.version_option(__version__, '--version', '-V')
def cli():
    """
    COLIN -- Container Linter
    """
    pass


@click.command(name="check",
               context_settings=CONTEXT_SETTINGS)
@click.argument('TARGET', type=click.STRING)
@click.option('--parent-target', type=click.STRING,
              help="Parent target")
@click.option('--ruleset', '-r', type=click.STRING, envvar='COLIN_RULESET',
              help="Select a predefined ruleset (e.g. fedora).")
@click.option('--ruleset-file', '-f', type=click.File(mode='r'),
              help="Path to a file to use for validation "
                   "(by default they are placed in /usr/share/colin/rulesets).")
@click.option('--debug', default=False, is_flag=True,
              help="Enable debugging mode (debugging logs, full tracebacks).")
@click.option('--json', type=click.File(mode='w'),
              help="File to save the output as json to.")
@click.option('--stat', is_flag=True,
              help="Print statistics instead of full results.")
@click.option('--skip', '-s', multiple=True, type=click.STRING,
              help="Name of the check to skip. (this option is repeatable)")
@click.option('--tag', '-t', multiple=True, type=click.STRING,
              help="Filter checks with the tag.")
@click.option('--verbose', '-v', is_flag=True,
              help="Verbose mode.")
@click.option('checks_paths', '--checks-path',
              type=click.Path(exists=True, dir_okay=True, file_okay=False),
              multiple=True, envvar=COLIN_CHECKS_PATH,
              help="Path to directory containing checks (default {}).".format(get_checks_paths()))
@click.option('--pull', is_flag=True,
              help="Pull the image from registry.")
@click.option('--target-type', type=click.STRING, default="image",
              help="Type of selected target (one of image, dockerfile, "
                   "ostree). For ostree, please specify image name and path like this: image@path")
@click.option('--timeout', type=click.INT,
              help="Timeout for each check in seconds. (default=600)")
@click.option('--insecure', is_flag=True, default=False,
              help="Pull from an insecure registry (HTTP or invalid TLS).")
def check(target, parent_target, ruleset, ruleset_file, debug, json, stat, skip, tag, verbose,
          checks_paths, target_type, timeout, pull, insecure):
    """
    Check the image/dockerfile (default).
    """
    if ruleset and ruleset_file:
        raise click.BadOptionUsage(
            "Options '--ruleset' and '--file-ruleset' cannot be used together.")

    if json and not os.path.isdir(os.path.dirname(os.path.realpath(json.name))):
        raise click.BadOptionUsage(
            "Parent directory for the json output file does not exist.")

    try:
        if not debug:
            logging.basicConfig(stream=six.StringIO())

        log_level = _get_log_level(debug=debug,
                                   verbose=verbose)
        results = run(
            target=target,
            parent_target=parent_target,
            ruleset_name=ruleset,
            ruleset_file=ruleset_file,
            logging_level=log_level,
            tags=tag,
            pull=pull,
            checks_paths=checks_paths,
            target_type=target_type,
            timeout=timeout,
            insecure=insecure,
            skips=skip
        )
        _print_results(results=results, stat=stat, verbose=verbose)

        if json:
            results.save_json_to_file(file=json)

        if not results.ok:
            sys.exit(1)
        elif results.fail:
            sys.exit(3)

    except ColinException as ex:
        logger.error("An error occurred: %r", ex)
        if debug:
            raise
        else:
            raise click.ClickException(str(ex))
    except Exception as ex:
        logger.error("An error occurred: %r", ex)
        if debug:
            raise
        else:
            raise click.ClickException(str(ex))


@click.command(name="list-checks",
               context_settings=CONTEXT_SETTINGS)
@click.option('--ruleset', '-r', type=click.STRING, envvar='COLIN_RULESET',
              help="Select a predefined ruleset (e.g. fedora).")
@click.option('--ruleset-file', '-f', type=click.File(mode='r'),
              help="Path to a file to use for validation "
                   "(by default they are placed in /usr/share/colin/rulesets).")
@click.option('--debug', default=False, is_flag=True,
              help="Enable debugging mode (debugging logs, full tracebacks).")
@click.option('--json', type=click.File(mode='w'),
              help="File to save the output as json to.")
@click.option('-s', '--skip', multiple=True, type=click.STRING,
              help="Name of the check to skip. (this option is repeatable)")
@click.option('--tag', '-t', multiple=True, type=click.STRING,
              help="Filter checks with the tag.")
@click.option('--verbose', '-v', is_flag=True,
              help="Verbose mode.")
@click.option('checks_paths', '--checks-path',
              type=click.Path(exists=True, dir_okay=True, file_okay=False),
              multiple=True, envvar=COLIN_CHECKS_PATH,
              help="Path to directory containing checks (default {}).".format(get_checks_paths()))
def list_checks(ruleset, ruleset_file, debug, json, skip, tag, verbose, checks_paths):
    """
    Print the checks.
    """
    if ruleset and ruleset_file:
        raise click.BadOptionUsage(
            "Options '--ruleset' and '--file-ruleset' cannot be used together.")

    try:
        if not debug:
            logging.basicConfig(stream=six.StringIO())

        log_level = _get_log_level(debug=debug,
                                   verbose=verbose)
        checks = get_checks(ruleset_name=ruleset,
                            ruleset_file=ruleset_file,
                            logging_level=log_level,
                            tags=tag,
                            checks_paths=checks_paths,
                            skips=skip)
        _print_checks(checks=checks)

        if json:
            AbstractCheck.save_checks_to_json(file=json, checks=checks)
    except ColinException as ex:
        logger.error("An error occurred: %r", ex)
        if debug:
            raise
        else:
            raise click.ClickException(str(ex))
    except Exception as ex:
        logger.error("An error occurred: %r", ex)
        if debug:
            raise
        else:
            raise click.ClickException(str(ex))


@click.command(name="list-rulesets",
               context_settings=CONTEXT_SETTINGS)
@click.option('--debug', default=False, is_flag=True,
              help="Enable debugging mode (debugging logs, full tracebacks).")
def list_rulesets(debug):
    """
    List available rulesets.
    """
    try:
        rulesets = get_rulesets()
        max_len = max([len(r[0]) for r in rulesets])
        for r in rulesets:
            click.echo('{0: <{1}} ({2})'.format(r[0], max_len, r[1]))
    except Exception as ex:
        logger.error("An error occurred: %r", ex)
        if debug:
            raise
        else:
            raise click.ClickException(str(ex))


@click.command(name="info",
               context_settings=CONTEXT_SETTINGS)
def info():
    """
    Show info about colin and its dependencies.
    """
    installation_path = os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))

    click.echo("colin {} {}".format(__version__, installation_path))
    click.echo("colin-cli {}\n".format(os.path.realpath(__file__)))

    # click.echo(get_version_of_the_python_package(module=conu))

    rpm_installed = is_rpm_installed()
    click.echo(get_version_msg_from_the_cmd(package_name="podman", use_rpm=rpm_installed))
    click.echo(get_version_msg_from_the_cmd(package_name="skopeo", use_rpm=rpm_installed))
    click.echo(get_version_msg_from_the_cmd(package_name="ostree",
                                            use_rpm=rpm_installed,
                                            max_lines_of_the_output=3))


cli.add_command(check)
cli.add_command(list_checks)
cli.add_command(list_rulesets)
cli.add_command(info)
cli.set_default_command(check)


def _print_results(results, stat=False, verbose=False):
    """
    Prints the results to the stdout

    :type verbose: bool
    :param results: generator of results
    :param stat: if True print stat instead of full output
    """
    results.generate_pretty_output(stat=stat,
                                   verbose=verbose,
                                   output_function=click.secho)


def _print_checks(checks):
    if not checks:
        click.echo("No check found.")
        return
    for check in checks:
        click.echo(str(check))


def _get_log_level(debug, verbose):
    if debug:
        return logging.DEBUG
    elif verbose:
        return logging.INFO
    else:
        return logging.WARNING


if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import re

from .abstract_check import ImageAbstractCheck
from ..result import CheckResult


class EnvCheck(ImageAbstractCheck):

    def __init__(self, message, description, reference_url, tags, env_var, required,
                 value_regex=None):
        super(EnvCheck, self) \
            .__init__(message, description, reference_url, tags)
        self.env_var = env_var
        self.required = required
        self.value_regex = value_regex

    def check(self, target):
        env_vars = target.config_metadata["Env"]

        env_vars_dict = {}
        if env_vars:
            for key_value in env_vars:
                key, value = key_value.split("=")
                env_vars_dict[key] = value
            present = self.env_var in env_vars_dict
        else:
            present = False

        if present:

            if self.required and not self.value_regex:
                passed = True
            elif self.value_regex:
                pattern = re.compile(self.value_regex)
                passed = bool(pattern.match(env_vars_dict[self.env_var]))
            else:
                passed = False

        else:
            passed = not self.required

        return CheckResult(ok=passed,
                           description=self.description,
                           message=self.message,
                           reference_url=self.reference_url,
                           check_name=self.name,
                           logs=[])



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import time

import pytest

from colin.core.exceptions import ColinException
from colin.utils.cmd_tools import exit_after
from colin.utils.cmd_tools import retry


@exit_after(1)
def fast_fce():
    pass


@exit_after(1)
def slow_fce():
    time.sleep(2)


@exit_after(1)
def bad_fce():
    raise ColinException("Error")


def test_timeout_fast_fce():
    fast_fce()


def test_timeout_slow_fce():
    with pytest.raises(TimeoutError):
        slow_fce()


def test_timeout_bad_fce():
    with pytest.raises(ColinException):
        bad_fce()


def test_timeout_dirrect():
    with pytest.raises(TimeoutError):
        exit_after(1)(time.sleep)(2)


COUNTER = 0


def raise_exception():
    global COUNTER
    COUNTER += 1

    raise Exception('I am bad function')


def test_no_retry_for_success():
    global COUNTER
    COUNTER = 0

    @retry(5, 0)
    def always_success():
        global COUNTER
        COUNTER = COUNTER + 1

        return 42

    assert always_success() == 42
    assert COUNTER == 1


def test_retry_with_exception():
    global COUNTER
    COUNTER = 0

    @retry(5, 0)
    def always_raise_exception():
        raise_exception()

    with pytest.raises(Exception) as ex:
        always_raise_exception()

    assert str(ex.value) == 'I am bad function'
    assert COUNTER == 5


def test_wrong_parameter():
    with pytest.raises(ValueError) as ex:
        retry(-1, 1)
    assert str(ex.value) == 'retry_count have to be positive'

    with pytest.raises(ValueError) as ex:
        retry(0, 1)
    assert str(ex.value) == 'retry_count have to be positive'

    @retry(5, -1)
    def fail_negative_sleep():
        raise_exception()

    with pytest.raises(ValueError) as ex:
        fail_negative_sleep()
    assert str(ex.value) == 'sleep length must be non-negative'


def test_retry_with_sleep():
    global COUNTER
    COUNTER = 0

    @retry(4, .5)
    def fail_and_sleep():
        raise_exception()

    time_start = time.time()
    with pytest.raises(Exception) as ex:
        fail_and_sleep()
    time_end = time.time()

    assert str(ex.value) == 'I am bad function'
    assert COUNTER == 4

    # there are 3 sleeps between 4 delays
    assert time_end - time_start >= 1.5
    # there were not 4 sleeps
    assert time_end - time_start < 4


def test_recover_after_few_failures():
    global COUNTER
    COUNTER = 0

    @retry(5, 0)
    def sleep_like_a_baby():
        global COUNTER
        if COUNTER < 3:
            COUNTER += 1
            raise Exception("sleeping")
        return []

    assert sleep_like_a_baby() == []
    assert COUNTER == 3



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import logging

import pytest

import colin
from colin.checks.labels import RunOrUsageLabelCheck
from colin.core.target import DockerfileTarget
from tests.conftest import get_target, LABELS_IMAGE_PARENT, build_image_if_not_exists


@pytest.fixture()
def empty_ruleset():
    return {
        "version": "1",
        "name": "Laughing out loud ruleset",
        "description": "This set of checks is required to pass because we said it",
        "contact_email": "forgot-to-reply@example.nope",
        "checks": [
        ]
    }


def get_results_from_colin_labels_image(target, parent_target=None):
    return colin.run(target=target.target_name,
                     parent_target=parent_target,
                     target_type=target.target_type,
                     ruleset_name="fedora",
                     logging_level=logging.DEBUG, pull=False)


def test_colin_image(target_label):
    assert get_results_from_colin_labels_image(target=target_label)


@pytest.mark.parametrize("same_parent_target", [None, True, False])
def test_labels_in_image(target_label, same_parent_target):
    parent = None
    if same_parent_target:
        parent = target_label.target_name
    elif same_parent_target is False:
        if not isinstance(target_label, DockerfileTarget):
            build_image_if_not_exists(LABELS_IMAGE_PARENT)
            parent_gen = get_target(LABELS_IMAGE_PARENT, target_label.target_type)
            parent_target = next(parent_gen)
            parent = parent_target.target_name

    result = get_results_from_colin_labels_image(target=target_label, parent_target=parent)
    assert result
    expected_dict = {"maintainer_label": "PASS",
                     "name_label": "PASS",
                     "com.redhat.component_label": "PASS",
                     "summary_label": "PASS",
                     "version_label": "PASS",
                     "run_or_usage_label": "PASS",
                     "release_label": "FAIL",
                     "architecture_label": "FAIL",
                     "url_label": "PASS",
                     "help_label": "FAIL",
                     "build-date_label": "FAIL",
                     "distribution-scope_label": "FAIL",
                     "vcs-ref_label": "FAIL",
                     "vcs-type_label": "FAIL",
                     "description_label": "PASS",
                     "io.k8s.description_label": "PASS",
                     "vcs-url_label": "FAIL",
                     "help_file_or_readme": "FAIL",
                     "inherited_labels": "PASS",
                     # "cmd_or_entrypoint": "PASS",
                     # "no_root": "FAIL",
                     }
    if same_parent_target:
        expected_dict['inherited_labels'] = "FAIL"

    if isinstance(target_label, DockerfileTarget):
        expected_dict.update({'description_or_io.k8s.description_label': 'PASS',
                              'from_tag_not_latest': 'FAIL',
                              'maintainer_deprecated': 'PASS'}
                             )
        del (expected_dict['help_file_or_readme'])
        del (expected_dict['run_or_usage_label'])
        del (expected_dict['inherited_labels'])
    labels_dict = {}
    for res in result.results:
        labels_dict[res.check_name] = res.status
    assert labels_dict == expected_dict


@pytest.mark.parametrize("labels, should_pass", [
    (["run"], True),
    (["usage"], False),
    (["run", "usage"], True),
    (["something", "different"], False),
    (["something", "completely", "different"], False),
])
def test_multiple_labels_check(labels, should_pass, target_label):
    check = RunOrUsageLabelCheck()
    check.labels = labels

    result = check.check(target_label)

    assert result.ok == should_pass



import re

from .abstract_check import ImageAbstractCheck, DockerfileAbstractCheck
from ..exceptions import ColinException


def check_label(labels, required, value_regex, target_labels):
    """
    Check if the label is required and match the regex

    :param labels: [str]
    :param required: bool (if the presence means pass or not)
    :param value_regex: str (using search method)
    :param target_labels: [str]
    :return: bool (required==True: True if the label is present and match the regex if specified)
                    (required==False: True if the label is not present)
    """
    present = target_labels is not None and not set(labels).isdisjoint(set(target_labels))

    if present:
        if required and not value_regex:
            return True
        elif value_regex:
            pattern = re.compile(value_regex)
            present_labels = set(labels) & set(target_labels)
            for l in present_labels:
                if not bool(pattern.search(target_labels[l])):
                    return False
            return True
        else:
            return False

    else:
        return not required


class NotLoadedCheck(DockerfileAbstractCheck, ImageAbstractCheck):

    def __init__(self, check_name, reason):
        self.name = check_name
        super(NotLoadedCheck, self).__init__(
            message="Check code '{}' {}.".format(check_name, reason),
            description="Did you set the right name in the ruleset file?",
            reference_url="",
            tags=[])

    def check(self, target):
        raise ColinException(self.message)



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from colin.core.checks.labels import DeprecatedLabelAbstractCheck
from colin.core.checks.fmf_check import FMFAbstractCheck


class ArchitectureLabelCapitalDeprecatedCheck(FMFAbstractCheck, DeprecatedLabelAbstractCheck):
    name = "architecture_label_capital_deprecated"


class BZComponentDeprecatedCheck(FMFAbstractCheck, DeprecatedLabelAbstractCheck):
    name = "bzcomponent_deprecated"


class InstallLabelCapitalDeprecatedCheck(FMFAbstractCheck, DeprecatedLabelAbstractCheck):
    name = "install_label_capital_deprecated"


class NameLabelCapitalDeprecatedCheck(FMFAbstractCheck, DeprecatedLabelAbstractCheck):
    name = "name_label_capital_deprecated"


class ReleaseLabelCapitalDeprecatedCheck(FMFAbstractCheck, DeprecatedLabelAbstractCheck):
    name = "release_label_capital_deprecated"


class UninstallLabelCapitalDeprecatedCheck(FMFAbstractCheck, DeprecatedLabelAbstractCheck):
    name = "uninstall_label_capital_deprecated"


class VersionLabelCapitalDeprecatedCheck(FMFAbstractCheck, DeprecatedLabelAbstractCheck):
    name = "version_label_capital_deprecated"



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import re

from .abstract_check import ImageAbstractCheck
from ..exceptions import ColinException
from ..result import CheckResult, FailedCheckResult


class CmdAbstractCheck(ImageAbstractCheck):

    def __init__(self, message, description, reference_url, tags, cmd, expected_output=None,
                 expected_regex=None,
                 substring=None):
        super(CmdAbstractCheck, self).__init__(message, description, reference_url, tags)
        self.cmd = cmd
        self.expected_output = expected_output
        self.expected_regex = expected_regex
        self.substring = substring

    def check(self, target):

        try:
            output = target.get_output(cmd=self.cmd)

            """
            except ConuException as ex:
                if str(ex).endswith("exit code 126") or str(ex).endswith("error: 127"):
                    return CheckResult(ok=False,
                                       description=self.description,
                                       message=self.message,
                                       reference_url=self.reference_url,
                                       check_name=self.name,
                                       logs=["exec: '{}': executable file not found in $PATH".format(
                                           self.cmd)])
                return FailedCheckResult(check=self,
                                         logs=[str(ex)])
            """
        except ColinException as ex:
            return FailedCheckResult(check=self,
                                     logs=[str(ex)])
        passed = True
        logs = ["Output:\n{}".format(output)]
        if self.substring is not None:
            substring_present = self.substring in output
            passed = passed and substring_present
            logs.append("{}: Substring '{}' is {}present in the output of the command '{}'." \
                        .format("ok" if substring_present else "nok",
                                self.substring,
                                "" if substring_present else "not ",
                                self.cmd))

        if self.expected_output is not None:
            expected_output = self.expected_output == output
            if expected_output:
                logs.append("ok: Output of the command '{}' "
                            "was as expected.".format(self.cmd))
            else:
                logs.append("nok: Output of the command '{}' "
                            "does not match the expected one: '{}'.".format(self.cmd,
                                                                            self.expected_output))

                passed = False

        if self.expected_regex is not None:
            pattern = re.compile(self.expected_regex)
            if pattern.match(output):
                logs.append("ok: Output of the command '{}' match the regex '{}'." \
                            .format(self.cmd, self.expected_regex))
            else:
                logs.append(
                    "nok: Output of the command '{}' does not match"
                    " the expected regex: '{}'.".format(self.cmd, self.expected_regex))

                passed = False

        return CheckResult(ok=passed,
                           description=self.description,
                           message=self.message,
                           reference_url=self.reference_url,
                           check_name=self.name,
                           logs=logs)



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import click
import six


class DefaultGroup(click.Group):
    """
    Allow to set default command for the group.
    """
    ignore_unknown_options = True

    def __init__(self, *args, **kwargs):
        default_command = kwargs.pop('default_command', None)
        super(DefaultGroup, self).__init__(*args, **kwargs)
        self.default_cmd_name = None
        if default_command is not None:
            self.set_default_command(default_command)

    def set_default_command(self, command):
        if isinstance(command, six.string_types):
            cmd_name = command
        else:
            cmd_name = command.name
            self.add_command(command)
        self.default_cmd_name = cmd_name

    def parse_args(self, ctx, args):
        if not args and self.default_cmd_name is not None:
            args.insert(0, self.default_cmd_name)
        return super(DefaultGroup, self).parse_args(ctx, args)

    def get_command(self, ctx, cmd_name):
        if cmd_name not in self.commands and self.default_cmd_name is not None:
            ctx.args0 = cmd_name
            cmd_name = self.default_cmd_name
        return super(DefaultGroup, self).get_command(ctx, cmd_name)

    def resolve_command(self, ctx, args):
        cmd_name, cmd, args = super(DefaultGroup, self).resolve_command(ctx, args)
        args0 = getattr(ctx, 'args0', None)
        if args0 is not None:
            args.insert(0, args0)
        return cmd_name, cmd, args



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import traceback

from .constant import CHECK_TIMEOUT
from .result import CheckResults, FailedCheckResult
from ..utils.cmd_tools import exit_after

logger = logging.getLogger(__name__)


def go_through_checks(target, checks, timeout=None):
    logger.debug("Going through checks.")
    results = _result_generator(target=target,
                                checks=checks,
                                timeout=timeout)
    return CheckResults(results=results)


def _result_generator(target, checks, timeout=None):
    try:
        for check in checks:
            logger.debug("Checking {}".format(check.name))
            try:
                timeout = timeout or check.timeout or CHECK_TIMEOUT
                logger.debug("Check timeout: {}".format(timeout))
                yield exit_after(timeout)(check.check)(target)
            except TimeoutError as ex:
                logger.warning(
                    "The check hit the timeout.")
                yield FailedCheckResult(check, logs=[str(ex)])
            except Exception as ex:
                tb = traceback.format_exc()
                logger.warning(
                    "There was an error while performing check: {}".format(tb))
                yield FailedCheckResult(check, logs=[str(ex)])
    finally:
        target.clean_up()



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from colin.core.checks.filesystem import FileCheck
from colin.core.checks.fmf_check import FMFAbstractCheck
from colin.core.checks.abstract_check import ImageAbstractCheck
from colin.core.result import CheckResult

logger = logging.getLogger(__name__)


class CmdOrEntrypointCheck(FMFAbstractCheck, ImageAbstractCheck):
    name = "cmd_or_entrypoint"

    def check(self, target):
        metadata = target.config_metadata["ContainerConfig"]
        cmd_present = "Cmd" in metadata and metadata["Cmd"]
        msg_cmd_present = "Cmd {}specified.".format("" if cmd_present else "not ")
        logger.debug(msg_cmd_present)

        entrypoint_present = "Entrypoint" in metadata and metadata["Entrypoint"]
        msg_entrypoint_present = "Entrypoint {}specified.".format(
            "" if entrypoint_present else "not ")
        logger.debug(msg_entrypoint_present)

        passed = cmd_present or entrypoint_present
        return CheckResult(ok=passed,
                           description=self.description,
                           message=self.message,
                           reference_url=self.reference_url,
                           check_name=self.name,
                           logs=[msg_cmd_present, msg_entrypoint_present])


class HelpFileOrReadmeCheck(FMFAbstractCheck, FileCheck):
    name = "help_file_or_readme"


class NoRootCheck(FMFAbstractCheck, ImageAbstractCheck):
    name = "no_root"

    def check(self, target):
        metadata = target.config_metadata
        root_present = "User" in metadata and metadata["User"] in ["", "0", "root"]

        return CheckResult(ok=not root_present,
                           description=self.description,
                           message=self.message,
                           reference_url=self.reference_url,
                           check_name=self.name,
                           logs=[])



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
This module is responsible for loading rulesets: reading from disk, parsing/validating
"""
import yaml
import logging

from ..exceptions import ColinRulesetException

logger = logging.getLogger(__name__)


def get_ruleset_struct_from_fileobj(fileobj):
    try:
        logger.debug("Loading ruleset from file '{}'.".format(fileobj.name))
        return RulesetStruct(yaml.safe_load(fileobj))
    except Exception as ex:
        msg = "Ruleset file '{}' cannot be loaded: {}".format(fileobj.name, ex)
        logger.error(msg)
        raise ColinRulesetException(msg)


def get_ruleset_struct_from_file(file_path):
    try:
        with open(file_path, "r") as fd:
            return get_ruleset_struct_from_fileobj(fd)
    except ColinRulesetException as ex:
        raise ex
    except Exception as ex:
        msg = "Ruleset '{}' cannot be loaded: {}".format(file_path, ex)

        logger.error(msg)
        raise ColinRulesetException(msg)


def nicer_get(di, required, *path):
    """
    this is a nicer way of doing dict.get()

    :param di: dict
    :param required: bool, raises an exc if value is not found, otherwise returns None
    :param path: list of str to navigate in the dict
    :return: your value
    """

    r = di
    for p in path:
        try:
            r = r[p]
        except KeyError:
            if required:
                logger.error("can't locate %s in ruleset dict, keys present: %s",
                             p, list(r.keys()))
                logger.debug("full dict = %s", r)
                raise ColinRulesetException("Validation error: can't locate %s in ruleset." % p)
            return
    return r


class CheckStruct(object):
    """
      {
        "name": "label_name",
        "tags": ["foo", "bar"],
        "additional_tags": ["baz"],
        "usable_targets": ["image", "dockerfile"],
      }
    """

    def __init__(self, check_dict):
        self.c = check_dict
        # TODO: validate the dict
        # TODO: get check class and merry them here

    def _get(self, required, *path):
        return nicer_get(self.c, required, *path)

    def __str__(self):
        return "%s" % self.name

    @property
    def import_name(self):
        """ module name + class, e.g. our.colin.checks.CustomCheck """
        return self._get(False, "import_name")

    @property
    def name(self):
        return self._get(True, "name")

    @property
    def tags(self):
        return self._get(False, "tags")

    @property
    def additional_tags(self):
        return self._get(False, "additional_tags")

    @property
    def usable_targets(self):
        return self._get(False, "usable_targets")

    @property
    def other_attributes(self):
        """ return dict with all other data except for the described above"""
        return {k: v for k, v in self.c.items() if
                k not in ["name", "names", "tags", "additional_tags", "usable_targets"]}


class RulesetStruct(object):
    """
    {
      "version": "1",
      "name": "Mandatory checks for Red Hat container images"
      "description": "This set of checks is required to pass on every container image ..."
      "contact_email": "cvp@redhat.com?"
      "checks": [{
        "name": "label_name",
        "tags": ["foo", "bar"],
        "additional_tags": ["baz"],
        "usable_targets": ["image", "dockerfile"],
      }, {}...
    ]}
    """

    def __init__(self, ruleset_dict):
        self.d = ruleset_dict
        # TODO: validate ruleset
        self._checks = None

    def _get(self, *path):
        return nicer_get(self.d, True, *path)

    def __str__(self):
        return "%s" % self.name

    @property
    def version(self):
        return self._get("version")

    @property
    def name(self):
        return self._get("name")

    @property
    def description(self):
        return self._get("description")

    @property
    def contact_email(self):
        return self._get("contact_email")

    @property
    def checks(self):
        if self._checks is None:
            self._checks = []
            for c in self._get("checks"):
                if "name" in c:
                    self._checks.append(CheckStruct(c))
                elif "names" in c:
                    for n in c["names"]:
                        new_check = dict(c)
                        del new_check["names"]
                        new_check["name"] = n
                        self._checks.append(CheckStruct(new_check))

        return self._checks



"""
Module with FMF abstract check class
"""

import logging
import inspect
import os

from .abstract_check import AbstractCheck
from ..fmf_extension import ExtendedTree

logger = logging.getLogger(__name__)


def receive_fmf_metadata(name, path, object_list=False):
    """
    search node identified by name fmfpath

    :param path: path to filesystem
    :param name: str - name as pattern to search - "/name" (prepended hierarchy item)
    :param object_list: bool, if true, return whole list of found items
    :return: Tree Object or list
    """
    output = {}
    fmf_tree = ExtendedTree(path)
    logger.debug("get FMF metadata for test (path:%s name=%s)", path, name)
    # ignore items with @ in names, to avoid using unreferenced items
    items = [x for x in fmf_tree.climb() if x.name.endswith("/" + name) and "@" not in x.name]
    if object_list:
        return items
    if len(items) == 1:
        output = items[0]
    elif len(items) > 1:
        raise Exception("There is more FMF test metadata for item by name:{}({}) {}".format(
            name, len(items), [x.name for x in items]))
    elif not items:
        raise Exception("Unable to get FMF metadata for: {}".format(name))
    return output


class FMFAbstractCheck(AbstractCheck):
    """
    Abstract class for checks and loading metadata from FMF format
    """
    metadata = None
    name = None
    fmf_metadata_path = None

    def __init__(self):
        """
        wraps parameters to COLIN __init__ method format
        """
        if not self.metadata:
            if not self.fmf_metadata_path:
                logger.info("setting self.fmf_metadata_path by class location."
                            " DO NOT use it in this way."
                            " Metadata are set in colin.core.loader (use proper path)")
                self.fmf_metadata_path = os.path.dirname(inspect.getfile(self.__class__))
            self.metadata = receive_fmf_metadata(name=self.name, path=self.fmf_metadata_path)
        master_class = super(FMFAbstractCheck, self)
        kwargs = {}
        try:
            # this is not available in python2, but second function is deprecated
            args_names = [argument for argument in
                          inspect.signature(master_class.__init__).parameters]
        except NameError:
            args_names = inspect.getargspec(master_class.__init__).args
        for arg in args_names:
            # copy all arguments from metadata.data to class __init__  kwargs
            try:
                kwargs[arg] = self.metadata.data[arg]
            except KeyError:
                pass
        try:
            master_class.__init__(**kwargs)
        except TypeError as error:
            logger.debug("missing argument (%s) in FMF metadata key (%s): %s",
                         error,
                         self.metadata.name,
                         self.metadata.data)



"""
Testing of low level interaction with images.
"""
import pytest


def test_file_is_present(target_busybox):
    assert target_busybox.file_is_present("/etc/passwd")
    assert not target_busybox.file_is_present("/oglogoblogologlo")
    with pytest.raises(IOError):
        target_busybox.file_is_present("/etc")


def test_labels_are_present(target_label):
    assert isinstance(target_label.labels, dict)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os

from setuptools import find_packages, setup

BASE_PATH = os.path.dirname(__file__)

# https://packaging.python.org/guides/single-sourcing-package-version/
version = {}
with open("./colin/version.py") as fp:
    exec(fp.read(), version)

long_description = ''.join(open('README.md').readlines())

setup(
    name='colin',
    version=version["__version__"],
    description="Tool to check generic rules/best-practices for containers/images/dockerfiles.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=['examples', 'tests']),
    install_requires=[
        'Click',
        'six',
        'dockerfile_parse',
        'fmf',
        'PyYAML'
    ],
    entry_points='''
        [console_scripts]
        colin=colin.cli.colin:cli
    ''',
    data_files=[("share/colin/rulesets/", ["rulesets/default.json",
                                           "rulesets/fedora.json"]),
                ("share/bash-completion/completions/", ["bash-complete/colin"])],
    license='GPLv3+',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Utilities',
    ],
    keywords='containers,sanity,linter',
    author='Red Hat',
    author_email='user-cont-team@redhat.com',
    url='https://github.com/user-cont/colin',
    package_data={'': ['*.fmf']},
    include_package_data=True,
)



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os

from .loader import (RulesetStruct, get_ruleset_struct_from_file,
                     get_ruleset_struct_from_fileobj)
from ..checks.check_utils import NotLoadedCheck
from ..constant import EXTS, RULESET_DIRECTORY, RULESET_DIRECTORY_NAME
from ..exceptions import ColinRulesetException
from ..loader import CheckLoader
from ..target import is_compatible

logger = logging.getLogger(__name__)


class Ruleset(object):

    def __init__(self, ruleset_name=None, ruleset_file=None, ruleset=None, checks_paths=None):
        """
        Load ruleset for colin.

        :param ruleset_name: str (name of the ruleset file (without any file extension), default is
                             "default"
        :param ruleset_file: fileobj instance holding ruleset configuration
        :param ruleset: dict, content of a ruleset file
        :param checks_paths: list of str, directories where the checks are present
        """
        self.check_loader = CheckLoader(get_checks_paths(checks_paths))
        if ruleset:
            self.ruleset_struct = RulesetStruct(ruleset)
        elif ruleset_file:
            self.ruleset_struct = get_ruleset_struct_from_fileobj(ruleset_file)
        else:
            logger.debug("Loading ruleset with the name '{}'.".format(ruleset_name))
            ruleset_path = get_ruleset_file(ruleset=ruleset_name)
            self.ruleset_struct = get_ruleset_struct_from_file(ruleset_path)
        if self.ruleset_struct.version not in ["1", 1]:
            raise ColinRulesetException("colin accepts only ruleset version '1'. You provided %r"
                                        % self.ruleset_struct.version)

    def get_checks(self, target_type, tags=None, skips=None):
        """
        Get all checks for given type/tags.

        :param skips: list of str
        :param target_type: TargetType class
        :param tags: list of str
        :return: list of check instances
        """
        skips = skips or []
        result = []
        for check_struct in self.ruleset_struct.checks:
            if check_struct.name in skips:
                continue

            logger.debug("Processing check_struct {}.".format(check_struct))

            usable_targets = check_struct.usable_targets
            if target_type and usable_targets \
                    and target_type.get_compatible_check_class().check_type not in usable_targets:
                logger.info("Skipping... Target type does not match.")
                continue

            if check_struct.import_name:
                check_class = self.check_loader.import_class(check_struct.import_name)
            else:
                try:
                    check_class = self.check_loader.mapping[check_struct.name]
                except KeyError:
                    logger.error("Check %s was not found -- it can't be loaded", check_struct.name)
                    raise ColinRulesetException(
                        "Check {} can't be loaded, we couldn't find it.".format(check_struct.name))
            check_instance = check_class()

            if check_struct.tags:
                logger.info("Overriding check's tags %s with the one defined in ruleset: %s",
                            check_instance.tags, check_struct.tags)
                check_instance.tags = check_struct.tags[:]
            if check_struct.additional_tags:
                logger.info("Adding additional tags: %s", check_struct.additional_tags)
                check_instance.tags += check_struct.additional_tags

            if not is_compatible(target_type=target_type, check_instance=check_instance):
                logger.error(
                    "Check '{}' not compatible with the target type: {}".format(
                        check_instance.name, target_type.get_compatible_check_class().check_type))
                raise ColinRulesetException(
                    "Check {} can't be used for target type {}".format(
                        check_instance, target_type.get_compatible_check_class().check_type))

            if tags:
                if not set(tags) < set(check_instance.tags):
                    logger.debug(
                        "Check '{}' not passed the tag control: {}".format(check_instance.name,
                                                                           tags))
                    continue

            # and finally, attach attributes from ruleset to the check instance
            for k, v in check_struct.other_attributes.items():
                # yes, this overrides things; yes, users may easily and severely broke their setup
                setattr(check_instance, k, v)

            result.append(check_instance)
            logger.debug("Check instance {} added.".format(check_instance.name))

        return result


def get_checks_paths(checks_paths=None):
    """
    Get path to checks.

    :param checks_paths: list of str, directories where the checks are present
    :return: list of str (absolute path of directory with checks)
    """
    p = os.path.join(__file__, os.pardir, os.pardir, os.pardir, "checks")
    p = os.path.abspath(p)
    # let's utilize the default upstream checks always
    if checks_paths:
        p += [os.path.abspath(x) for x in checks_paths]
    return [p]


def get_ruleset_file(ruleset=None):
    """
    Get the ruleset file from name

    :param ruleset: str
    :return: str
    """
    ruleset = ruleset or "default"

    ruleset_dirs = get_ruleset_dirs()
    for ruleset_directory in ruleset_dirs:
        possible_ruleset_files = [os.path.join(ruleset_directory, ruleset + ext) for ext in EXTS]

        for ruleset_file in possible_ruleset_files:
            if os.path.isfile(ruleset_file):
                logger.debug("Ruleset file '{}' found.".format(ruleset_file))
                return ruleset_file

    logger.warning("Ruleset with the name '{}' cannot be found at '{}'."
                   .format(ruleset, ruleset_dirs))
    raise ColinRulesetException("Ruleset with the name '{}' cannot be found.".format(ruleset))


def get_ruleset_dirs():
    """
    Get the directory with ruleset files
    First directory to check:  ./rulesets
    Second directory to check:  $HOME/.local/share/colin/rulesets
    Third directory to check: /usr/local/share/colin/rulesets
    :return: str
    """

    ruleset_dirs = []

    cwd_rulesets = os.path.join(".", RULESET_DIRECTORY_NAME)
    if os.path.isdir(cwd_rulesets):
        logger.debug("Ruleset directory found in current directory ('{}').".format(cwd_rulesets))
        ruleset_dirs.append(cwd_rulesets)

    if "VIRTUAL_ENV" in os.environ:
        venv_local_share = os.path.join(os.environ["VIRTUAL_ENV"],
                                        RULESET_DIRECTORY)
        if os.path.isdir(venv_local_share):
            logger.debug("Virtual env ruleset directory found ('{}').".format(venv_local_share))
            ruleset_dirs.append(venv_local_share)

    local_share = os.path.join(os.path.expanduser("~"),
                               ".local",
                               RULESET_DIRECTORY)
    if os.path.isdir(local_share):
        logger.debug("Local ruleset directory found ('{}').".format(local_share))
        ruleset_dirs.append(local_share)

    usr_local_share = os.path.join("/usr/local", RULESET_DIRECTORY)
    if os.path.isdir(usr_local_share):
        logger.debug("Global ruleset directory found ('{}').".format(usr_local_share))
        ruleset_dirs.append(usr_local_share)

    if not ruleset_dirs:
        msg = "Ruleset directory cannot be found."
        logger.warning(msg)
        raise ColinRulesetException(msg)

    return ruleset_dirs


def get_rulesets():
    """"
    Get available rulesets.
    """
    rulesets_dirs = get_ruleset_dirs()
    ruleset_files = []
    for rulesets_dir in rulesets_dirs:
        for f in os.listdir(rulesets_dir):
            for ext in EXTS:
                file_path = os.path.join(rulesets_dir, f)
                if os.path.isfile(file_path) and f.lower().endswith(ext):
                    ruleset_files.append((f[:-len(ext)], file_path))
    return ruleset_files



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import logging

from .abstract_check import ImageAbstractCheck
from ..result import CheckResult

logger = logging.getLogger(__name__)


class FileCheck(ImageAbstractCheck):
    """ Check presence of files; w/o mounting the whole FS """

    def __init__(self, message, description, reference_url, tags, files, all_must_be_present):
        super(FileCheck, self) \
            .__init__(message, description, reference_url, tags)
        self.files = files
        self.all_must_be_present = all_must_be_present

    def _handle_image(self, target):
        passed = self.all_must_be_present

        logs = []
        for f in self.files:
            try:
                f_present = target.file_is_present(f)
                logs.append("File '{}' is {}present."
                            .format(f, "" if f_present else "not "))
            except IOError as ex:
                logger.info("File %s is not present, ex: %s", f, ex)
                f_present = False
                logs.append("File {} is not present.".format(f))
            if self.all_must_be_present:
                passed = f_present and passed
            else:
                passed = f_present or passed

        for log in logs:
            logger.debug(log)

        return CheckResult(ok=passed,
                           description=self.description,
                           message=self.message,
                           reference_url=self.reference_url,
                           check_name=self.name,
                           logs=logs)

    def _handle_container(self, target):
        passed = self.all_must_be_present
        cont = target.instance

        logs = []
        for f in self.files:
            cmd = ["/bin/ls", "-1", f]
            try:
                f_present = cont.execute(cmd)
                logs.append("File '{}' is {}present."
                            .format(f, "" if f_present else "not "))
            except Exception as ex:
                logger.info("File %s is not present, ex: %s", f, ex)
                f_present = False
                logs.append("File {} is not present.".format(f))
            if self.all_must_be_present:
                passed = f_present and passed
            else:
                passed = f_present or passed

        return CheckResult(ok=passed,
                           description=self.description,
                           message=self.message,
                           reference_url=self.reference_url,
                           check_name=self.name,
                           logs=logs)

    def check(self, target):
        return self._handle_image(target)

# class FileSystemCheck(ImageAbstractCheck):
#     """ check for presence of files using `docker save` """
#
#     def __init__(self, message, description, reference_url, tags, files, all_must_be_present):
#         super(FileSystemCheck, self) \
#             .__init__(message, description, reference_url, tags)
#         self.files = files
#         self.all_must_be_present = all_must_be_present
#
#     def check(self, target):
#         try:
#             with target.instance.mount() as fs:
#                 passed = self.all_must_be_present
#
#                 logs = []
#                 for f in self.files:
#                     try:
#                         f_present = fs.file_is_present(f)
#                         logs.append("File '{}' is {}present."
#                                     .format(f, "" if f_present else "not "))
#                     except IOError as ex:
#                         f_present = False
#                         logs.append("Error: {}".format(str(ex)))
#
#                     if self.all_must_be_present:
#                         passed = f_present and passed
#                     else:
#                         passed = f_present or passed
#
#                 return CheckResult(ok=passed,
#                                    description=self.description,
#                                    message=self.message,
#                                    reference_url=self.reference_url,
#                                    check_name=self.name,
#                                    logs=logs)
#         except Exception as ex:
#             raise ColinException("There was an error while operating on filesystem of {}: {}"
#                                  .format(target.instance, str(ex)))



"""
This is a temporary module to support unpriv way of interacting with container images.

It will be migrated to conu sooner or later.
"""
import logging

logger = logging.getLogger(__name__)


class ImageName(object):
    """ parse image references and access their components easily """

    def __init__(self, registry=None, namespace=None, repository=None, tag=None, digest=None):
        self.registry = registry
        self.namespace = namespace
        self.repository = repository
        self.digest = digest
        self.tag = tag

    @classmethod
    def parse(cls, image_name):
        """
        Get the instance of ImageName from the string representation.

        :param image_name: str (any possible form of image name)
        :return: ImageName instance
        """
        result = cls()

        # registry.org/namespace/repo:tag
        s = image_name.split('/', 2)

        if len(s) == 2:
            if '.' in s[0] or ':' in s[0]:
                result.registry = s[0]
            else:
                result.namespace = s[0]
        elif len(s) == 3:
            result.registry = s[0]
            result.namespace = s[1]
        result.repository = s[-1]

        try:
            result.repository, result.digest = result.repository.rsplit("@", 1)
        except ValueError:
            try:
                result.repository, result.tag = result.repository.rsplit(":", 1)
            except ValueError:
                result.tag = "latest"

        return result

    def __str__(self):
        return "Image: registry='{}' namespace='{}' " \
               "repository='{}' tag='{}' digest='{}'".format(self.registry,
                                                             self.namespace,
                                                             self.repository,
                                                             self.tag,
                                                             self.digest)

    @property
    def name(self):
        """
        Get the string representation of the image
        (registry, namespace, repository and digest together).

        :return: str
        """
        name_parts = []
        if self.registry:
            name_parts.append(self.registry)

        if self.namespace:
            name_parts.append(self.namespace)

        if self.repository:
            name_parts.append(self.repository)
        name = "/".join(name_parts)

        if self.digest:
            name += "@{}".format(self.digest)
        elif self.tag:
            name += ":{}".format(self.tag)

        return name



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

RULESET_DIRECTORY_NAME = "rulesets"
RULESET_DIRECTORY = "share/colin/" + RULESET_DIRECTORY_NAME
EXTS = [".yaml", '.yml', ".json"]

PASSED = "PASS"
FAILED = "FAIL"
ERROR = "ERROR"

COLOURS = {
    PASSED: "green",
    FAILED: "red",
    ERROR: "red"
}

OUTPUT_CHARS = {
    PASSED: ".",
    FAILED: "x",
    ERROR: "#"
}

COLIN_CHECKS_PATH = "CHECKS_PATH"

CHECK_TIMEOUT = 10 * 60  # s



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import shutil

import colin.checks
from colin.core.loader import CheckLoader


def test_upstream_checks_can_be_loaded():
    """ check whether all upstream checks can be loaded """
    colin_checks_path = colin.checks.__file__
    colin_checks_dir = os.path.dirname(colin_checks_path)
    l = CheckLoader([colin_checks_dir])
    assert l.check_classes
    assert l.mapping["bzcomponent_deprecated"]
    assert l.mapping["architecture_label_capital_deprecated"]
    assert l.mapping["bzcomponent_deprecated"]
    assert l.mapping["install_label_capital_deprecated"]
    assert l.mapping["name_label_capital_deprecated"]
    assert l.mapping["release_label_capital_deprecated"]
    assert l.mapping["uninstall_label_capital_deprecated"]
    assert l.mapping["version_label_capital_deprecated"]
    assert l.mapping["architecture_label"]
    assert l.mapping["authoritative_source-url_label"]
    assert l.mapping["build-date_label"]
    assert l.mapping["com.redhat.build-host_label"]
    assert l.mapping["com.redhat.component_label"]
    assert l.mapping["description_label"]
    assert l.mapping["description_or_io.k8s.description_label"]
    assert l.mapping["distribution-scope_label"]
    assert l.mapping["help_label"]
    assert l.mapping["io.k8s.description_label"]
    assert l.mapping["io.k8s.display-name_label"]
    assert l.mapping["maintainer_label"]
    assert l.mapping["name_label"]
    assert l.mapping["release_label"]
    assert l.mapping["summary_label"]
    assert l.mapping["url_label"]
    assert l.mapping["run_or_usage_label"]
    assert l.mapping["vcs-ref_label"]
    assert l.mapping["vcs-type_label"]
    assert l.mapping["vcs-url_label"]
    assert l.mapping["vendor_label"]
    assert l.mapping["version_label"]
    assert l.mapping["cmd_or_entrypoint"]
    assert l.mapping["help_file_or_readme"]
    assert l.mapping["no_root"]
    # assert l.mapping["shell_runnable"]   # FIXME: commented out before move to podman
    assert l.mapping["from_tag_not_latest"]
    assert l.mapping["maintainer_deprecated"]


def test_loading_custom_check(tmpdir):
    tests_dir = os.path.dirname(os.path.dirname(__file__))
    a_check_dir = os.path.join(tests_dir, "data", "a_check")
    shutil.copytree(a_check_dir, str(tmpdir.join("a_check")))
    l = CheckLoader([str(tmpdir)])
    assert len(l.check_classes) == 3
    assert l.mapping["a-peter-file-check"]
    assert l.mapping["this-is-a-funky-check"]


def test_import_class():
    l = CheckLoader([])
    check_name = "ArchitectureLabelCheck"
    c = l.import_class("colin.checks.labels.%s" % check_name)
    assert c.name == "architecture_label"



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This piece of code searches for python code on specific path and
loads AbstractCheck classes from it.
"""

import inspect
import logging
import os
from importlib import import_module
from importlib.util import module_from_spec
from importlib.util import spec_from_file_location

from ..core.checks.fmf_check import receive_fmf_metadata, FMFAbstractCheck

logger = logging.getLogger(__name__)


def path_to_module(path):
    if path.endswith(".py"):
        path = path[:-3]
    cleaned_path = path.replace(".", "").replace("-", "_")
    path_comps = cleaned_path.split(os.sep)[-2:]
    import_name = ".".join(path_comps)
    if import_name[0] == ".":
        import_name = import_name[1:]
    return import_name


def _load_module(path):
    module_name = path_to_module(path)
    logger.debug("Will try to load selected file as module '%s'.", module_name)

    s = spec_from_file_location(module_name, path)
    m = module_from_spec(s)
    s.loader.exec_module(m)
    return m


def should_we_load(kls):
    """ should we load this class as a check? """
    # we don't load abstract classes
    if kls.__name__.endswith("AbstractCheck"):
        return False
    # and we only load checks
    if not kls.__name__.endswith("Check"):
        return False
    mro = kls.__mro__
    # and the class needs to be a child of AbstractCheck
    for m in mro:
        if m.__name__ == "AbstractCheck":
            return True
    return False


def load_check_classes_from_file(path):
    logger.debug("Getting check(s) from the file '{}'.".format(path))
    m = _load_module(path)

    check_classes = []
    for _, obj in inspect.getmembers(m, inspect.isclass):
        if should_we_load(obj):
            if issubclass(obj, FMFAbstractCheck):
                node_metadata = receive_fmf_metadata(name=obj.name, path=os.path.dirname(path))
                obj.metadata = node_metadata
            check_classes.append(obj)
            # Uncomment when debugging this code.
            logger.debug("Check class '%s' loaded, module: '%s'", obj.__name__, obj.__module__)
    return check_classes


class CheckLoader(object):
    """
    find recursively all checks on a given path
    """

    def __init__(self, checks_paths):
        """
        :param checks_paths: list of str, directories where the checks are present
        """
        logger.debug("Will load checks from paths '%s'.", checks_paths)
        for p in checks_paths:
            if os.path.isfile(p):
                raise RuntimeError("Provided path %s is not a directory." % p)
        self._check_classes = None
        self._mapping = None
        self.paths = checks_paths

    def obtain_check_classes(self):
        """ find children of AbstractCheck class and return them as a list """
        check_classes = set()
        for path in self.paths:
            for root, _, files in os.walk(path):
                for fi in files:
                    if not fi.endswith(".py"):
                        continue
                    path = os.path.join(root, fi)
                    check_classes = check_classes.union(set(
                        load_check_classes_from_file(path)))
        return list(check_classes)

    def import_class(self, import_name):
        """
        import selected class

        :param import_name, str, e.g. some.module.MyClass
        :return the class
        """
        module_name, class_name = import_name.rsplit(".", 1)
        mod = import_module(module_name)
        check_class = getattr(mod, class_name)
        self.mapping[check_class.name] = check_class
        logger.info("successfully loaded class %s", check_class)
        return check_class

    @property
    def check_classes(self):
        if self._check_classes is None:
            self._check_classes = self.obtain_check_classes()
        return self._check_classes

    @property
    def mapping(self):
        if self._mapping is None:
            self._mapping = {}
            for c in self.check_classes:
                self._mapping[c.name] = c
        return self._mapping



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from .check_runner import go_through_checks
from .ruleset.ruleset import Ruleset
from .target import Target

logger = logging.getLogger(__name__)


def run(
    target,
    target_type,
    parent_target=None,
    tags=None,
    ruleset_name=None,
    ruleset_file=None,
    ruleset=None,
    logging_level=logging.WARNING,
    checks_paths=None,
    pull=None,
    insecure=False,
    skips=None,
    timeout=None,
):
    """
    Runs the sanity checks for the target.

    :param timeout: timeout per-check (in seconds)
    :param skips: name of checks to skip
    :param target: str (image name, ostree or dockertar)
                    or ImageTarget
                    or path/file-like object for dockerfile
    :param parent_target: Target for parent image
    :param target_type: string, either image, dockerfile, dockertar
    :param tags: list of str (if not None, the checks will be filtered by tags.)
    :param ruleset_name: str (e.g. fedora; if None, default would be used)
    :param ruleset_file: fileobj instance holding ruleset configuration
    :param ruleset: dict, content of a ruleset file
    :param logging_level: logging level (default logging.WARNING)
    :param checks_paths: list of str, directories where the checks are present
    :param pull: bool, pull the image from registry
    :param insecure: bool, pull from an insecure registry (HTTP/invalid TLS)
    :return: Results instance
    """
    _set_logging(level=logging_level)
    logger.debug("Checking started.")

    parent = None
    if parent_target is not None and target_type != 'dockerfile':
        parent = Target.get_instance(
            target=parent_target,
            logging_level=logging_level,
            pull=pull,
            target_type=target_type,
            insecure=insecure,
        )

    target = Target.get_instance(
        target=target,
        parent_target=parent,
        logging_level=logging_level,
        pull=pull,
        target_type=target_type,
        insecure=insecure,
    )

    checks_to_run = _get_checks(
        target_type=target.__class__,
        tags=tags,
        ruleset_name=ruleset_name,
        ruleset_file=ruleset_file,
        ruleset=ruleset,
        checks_paths=checks_paths,
        skips=skips,
    )
    result = go_through_checks(target=target, checks=checks_to_run, timeout=timeout)
    return result


def get_checks(
    target_type=None,
    tags=None,
    ruleset_name=None,
    ruleset_file=None,
    ruleset=None,
    logging_level=logging.WARNING,
    checks_paths=None,
    skips=None,
):
    """
    Get the sanity checks for the target.

    :param skips: name of checks to skip
    :param target_type: TargetType enum
    :param tags: list of str (if not None, the checks will be filtered by tags.)
    :param ruleset_name: str (e.g. fedora; if None, default would be used)
    :param ruleset_file: fileobj instance holding ruleset configuration
    :param ruleset: dict, content of a ruleset file
    :param logging_level: logging level (default logging.WARNING)
    :param checks_paths: list of str, directories where the checks are present
    :return: list of check instances
    """
    _set_logging(level=logging_level)
    logger.debug("Finding checks started.")
    return _get_checks(
        target_type=target_type,
        tags=tags,
        ruleset_name=ruleset_name,
        ruleset_file=ruleset_file,
        ruleset=ruleset,
        checks_paths=checks_paths,
        skips=skips,
    )


def _get_checks(
    target_type,
    tags=None,
    ruleset_name=None,
    ruleset_file=None,
    ruleset=None,
    checks_paths=None,
    skips=None,
):
    ruleset = Ruleset(
        ruleset_name=ruleset_name,
        ruleset_file=ruleset_file,
        ruleset=ruleset,
        checks_paths=checks_paths,
    )
    return ruleset.get_checks(tags=tags, target_type=target_type, skips=skips)


def _set_logging(
        logger_name="colin",
        level=logging.INFO,
        handler_class=logging.StreamHandler,
        handler_kwargs=None,
        format='%(asctime)s.%(msecs).03d %(filename)-17s %(levelname)-6s %(message)s',
        date_format='%H:%M:%S'):
    """
    Set personal logger for this library.

    :param logger_name: str, name of the logger
    :param level: int, see logging.{DEBUG,INFO,ERROR,...}: level of logger and handler
    :param handler_class: logging.Handler instance, default is StreamHandler (/dev/stderr)
    :param handler_kwargs: dict, keyword arguments to handler's constructor
    :param format: str, formatting style
    :param date_format: str, date style in the logs
    """
    if level != logging.NOTSET:
        logger = logging.getLogger(logger_name)
        logger.setLevel(level)

        # do not readd handlers if they are already present
        if not [x for x in logger.handlers if isinstance(x, handler_class)]:
            handler_kwargs = handler_kwargs or {}
            handler = handler_class(**handler_kwargs)
            handler.setLevel(level)

            formatter = logging.Formatter(format, date_format)
            handler.setFormatter(formatter)
            logger.addHandler(handler)



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import functools
import logging
import subprocess
import threading
import time

try:
    import thread
except ImportError:
    import _thread as thread

logger = logging.getLogger(__name__)


def get_version_of_the_python_package(module):
    """
    Return the str containing the name version and package location.

    :param module: module to show info about
    :return: str 'name version path'
    """
    return "{} {} {}".format(getattr(module, "__name__", None),
                             getattr(module, "__version__", None),
                             getattr(module, "__path__", [None])[0])


def get_version_msg_from_the_cmd(package_name, cmd=None, use_rpm=None,
                                 max_lines_of_the_output=None):
    """
    Get str with the version (or string representation of the error).

    :param package_name: str
    :param cmd: str or [str] (defaults to [package_name, "--version"])
    :param use_rpm: True/False/None (whether to use rpm -q for getting a version)
    :param max_lines_of_the_output: use first n lines of the output
    :return: str
    """
    if use_rpm is None:
        use_rpm = is_rpm_installed()
    if use_rpm:
        rpm_version = get_rpm_version(package_name=package_name)
        if rpm_version:
            return "{} (rpm)".format(rpm_version)

    try:
        cmd = cmd or [package_name, "--version"]
        version_result = subprocess.run(cmd,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE)
        if version_result.returncode == 0:
            version_output = version_result.stdout.decode().rstrip()
            if max_lines_of_the_output:
                version_output = " ".join(version_output.split('\n')[:max_lines_of_the_output])
            return version_output

        else:
            return "{}: cannot get version with {}".format(package_name, cmd)
    except FileNotFoundError:
        return "{} not accessible!".format(package_name)


def get_rpm_version(package_name):
    """Get a version of the package with 'rpm -q' command."""
    version_result = subprocess.run(["rpm", "-q", package_name],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
    if version_result.returncode == 0:
        return version_result.stdout.decode().rstrip()
    else:
        return None


def is_rpm_installed():
    """Tests if the rpm command is present."""
    try:
        version_result = subprocess.run(["rpm", "--usage"],
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE)
        rpm_installed = not version_result.returncode
    except FileNotFoundError:
        rpm_installed = False
    return rpm_installed


def exit_after(s):
    """
    Use as decorator to exit process if
    function takes longer than s seconds.

    Direct call is available via exit_after(TIMEOUT_IN_S)(fce)(args).

    Inspired by https://stackoverflow.com/a/31667005
    """

    def outer(fn):
        def inner(*args, **kwargs):
            timer = threading.Timer(s, thread.interrupt_main)
            timer.start()
            try:
                result = fn(*args, **kwargs)
            except KeyboardInterrupt:
                raise TimeoutError("Function '{}' hit the timeout ({}s).".format(fn.__name__, s))
            finally:
                timer.cancel()
            return result

        return inner

    return outer


def retry(retry_count=5, delay=2):
    """
    Use as decorator to retry functions few times with delays

    Exception will be raised if last call fails

    :param retry_count: int could of retries in case of failures. It must be
                        a positive number
    :param delay: int delay between retries
    """
    if retry_count <= 0:
        raise ValueError("retry_count have to be positive")

    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            for i in range(retry_count, 0, -1):
                try:
                    return f(*args, **kwargs)
                except Exception:
                    if i <= 1:
                        raise
                time.sleep(delay)
        return wrapper
    return decorator
